propertyTypes = [
	{n: "Distribution", t: "g", r: []},
	{n: "DistributionType",      t: "e", r: [
		{l: "Exponential",   op: "exp",    tt: "Exponential distribution. Parameter <rate> holds the rate"},
		{l: "Deterministic", op: "det",    tt: "Deterministic distribution. Parameter <val> holds the value"},
		{l: "Uniform",       op: "unif",   tt: "Uniform distribution. Parameters <min, max> hold the extremes"},
		{l: "Erlang",        op: "erlang", tt: "Erlang distribution. Parameter <lambda> holds the rate, <k> the number of stages"}
	]},
	{n: "String", t: "s", r: {}},
	{n: "Int", t: "i", r: {}},
	{n: "QueuePolicyType", t: "e", r: [
		{l: "FCFS", op: "FCFS", tt: "First Come, First Served"},
		{l: "LCFS", op: "LCFS", tt: "Last Come, First Served"},
		{l: "SRO",  op: "SRO",  tt: "Service in Random Order"},
		{l: "SJF",  op: "SJF",  tt: "Shortest Jobs First"}
	]},
	{n: "RoutingPolicyType", t: "e", r: [
		{l: "SINGLE", op: "SINGLE", tt: "Single destination - usefull only if there is just one output connection"},
		{l: "PROB",   op: "PROB",   tt: "Probabilistic Ruting - probability is specified on the output arcs"},
		{l: "RANDOM", op: "RANDOM", tt: "Random: probabilistic with equal probability"},
		{l: "RR",     op: "RR",     tt: "Round Robin"},
		{l: "JSQ",    op: "JSQ",    tt: "Join the Shortest Queue"}
	]},
	{n: "Float", t: "f", r: {}}
]

ptMap = {};
for(i = 0; i < propertyTypes.length; i++) {
	ptMap[propertyTypes[i].n] = i;
}

propertyCollection = [
	{n: "Lambda", t: ptMap.Distribution, l: "Lambda", tt: "Inter Arrival Time Distribution", d: {DistributionType: "EXP", DistributionPars: "rate = 8"}},
	{n: "NumberOfJobs", t: ptMap.Int, l: "Number Of Jobs", tt: "Inject this number of jobs in the connected station", d: 1},
	{n: "D", t: ptMap.Distribution, l: "Service", tt: "Service Time Distribution", d: {DistributionType: "EXP", DistributionPars: "rate = 10"}},
	{n: "Q", t: ptMap.QueuePolicyType, l: "Queuing Policy", tt: "How to select the next job", d: "FCFS"},
	{n: "NumberServers", t: ptMap.Int, l: "Number Servers", tt: "Number of parallel servers", d: 1},
	{n: "Rt", t: ptMap.RoutingPolicyType, l: "Routing Policy", tt: "How to select the next destination", d: "RANDOM"},
	{n: "Prob", t: ptMap.Float, l: "Routing Probability", tt: "For probabilistic routing: probaiblity of moving across this arc", d: 1},
	{n: "DistributionType", t: ptMap.DistributionType, l: "Distribution", tt: "Type of Probability Distribution", d: "exp"},
	{n: "DistributionPars", t: ptMap.String, l: "Parameters", tt: "Comma separated list of the Parameters of the Probability Distribution", d: "rate = 1"},
];


pcMap = {};
for(i = 0; i < propertyCollection.length; i++) {
	pcMap[propertyCollection[i].n] = i;
}

propertyTypes[ptMap.Distribution].r = [pcMap.DistributionType, pcMap.DistributionPars];


elementTypes = [
    {t: "n", d: {sh: "g", hbh: 20, hbw: 20,	d: [
												{x:-8, y:10, sh: "t", hbh: 5.5, hbw: 12,		 fcol: "#880000", sz: 30, txt:"S", font:"Times", w:0, h:0},
												{x: 0, y: 0, sh: "p", d: "M20 5 L20 20 L-20 20 L-20 -20 L20 -20 L20 -5", scol: "#000000", swidth: 1.5}
    										   ] },
		ft: "arrival", tt: "Arrivals (Open Models)", l: "Arrivals", p: [
				pcMap.Lambda,
				pcMap.Rt
		]
											   },
    {t: "n", d: {sh: "g", hbh: 20, hbw: 20,	d: [
												{x:-10, y:10, sh: "t", hbh: 5.5, hbw: 12,		 fcol: "#007700", sz: 30, txt:"C", font:"Times", w:0, h:0},
												{x: 0, y: 0, sh: "p", d: "M20 5 L20 20 L-20 20 L-20 -20 L20 -20 L20 -5", scol: "#000000", swidth: 1.5}
    										   ] },
		ft: "closedInit", tt: "Initial Population (Closed Models)", l: "Closed Init", p: [
				pcMap.NumberOfJobs,
				pcMap.Rt
		]
											    },
    {t: "n", d: {sh: "g", hbh: 10, hbw: 22,	d: [
					    						{x: -10, y: 0, sh: "r", fcol: "#ffffff", scol: "#000000", swidth: 1.5, h: 20, w: 30},
					    						{x:  15, y: 0, sh: "c", fcol: "#ffffff", scol: "#000000", swidth: 1.5, sz: 20},
												{x: -2.5, y: 0, sh: "p", d: "M0 -10 L0 10", scol: "#000000", swidth: 1.5},
												{x: -10, y: 0, sh: "p", d: "M0 -10 L0 10", scol: "#000000", swidth: 1.5},
												{x: -17.5, y: 0, sh: "p", d: "M0 -10 L0 10", scol: "#000000", swidth: 1.5}
    										   ] },
		ft: "service", tt: "Single Server Queuing Station", l: "Single Server", p: [
				pcMap.D,
				pcMap.Q,
				pcMap.Rt
		]
											    },
    {t: "n", d: {sh: "g", hbh: 24, hbw: 20,	d: [
					    						{x: -10, y: 0, sh: "r", fcol: "#ffffff", scol: "#000000", swidth: 1.5, h: 20, w: 30},
					    						{x:  12.5, y: -20, sh: "c", fcol: "#ffffff", scol: "#000000", swidth: 1.5, sz: 15},
					    						{x:  12.5, y: -5, sh: "c", fcol: "#ffffff", scol: "#000000", swidth: 1.5, sz: 15},
					    						{x:  12.5, y: 20, sh: "c", fcol: "#ffffff", scol: "#000000", swidth: 1.5, sz: 15},
												{x:5, y:8, sh: "t", hbh: 5.5, hbw: 12,		 fcol: "#000000", sz: 20, txt:"...", font:"Arial", w:0, h:0},
												{x: -2.5, y: 0, sh: "p", d: "M0 -10 L0 10", scol: "#000000", swidth: 1.5},
												{x: -10, y: 0, sh: "p", d: "M0 -10 L0 10", scol: "#000000", swidth: 1.5},
												{x: -17.5, y: 0, sh: "p", d: "M0 -10 L0 10", scol: "#000000", swidth: 1.5}
    										   ] },
		ft: "multipleService", tt: "Multiple Server Queuing Station", l: "Multiple Server", p: [
				pcMap.D,
				pcMap.Q,
				pcMap.NumberServers,
				pcMap.Rt
		]
											    },
    {t: "n", d: {sh: "g", hbh: 24, hbw: 15,	d: [
					    						{x:  0, y: -20, sh: "c", fcol: "#ffffff", scol: "#000000", swidth: 1.5, sz: 15},
					    						{x:  0, y: -5, sh: "c", fcol: "#ffffff", scol: "#000000", swidth: 1.5, sz: 15},
					    						{x:  0, y: 20, sh: "c", fcol: "#ffffff", scol: "#000000", swidth: 1.5, sz: 15},
												{x:-7.5, y:8, sh: "t", hbh: 5.5, hbw: 12,		 fcol: "#000000", sz: 20, txt:"...", font:"Arial", w:0, h:0},
												{x: -17.5, y: 0, sh: "p", d: "M0 5 L10 20 M0 -5 L10 -20", scol: "#000000", swidth: 1.5},
												{x: 17.5, y: 0, sh: "p", d: "M0 5 L-10 20 M0 -5 L-10 -20", scol: "#000000", swidth: 1.5}
    										   ] },
		ft: "infiniteService", tt: "Infinite Server Queuing Station (Delay)", l: "Delay", p: [
				pcMap.D,
				pcMap.Rt
		]
											    },
    {t: "n", d: {sh: "g", hbh: 10, hbw: 22,	d: [
					    						{x: -10, y: 0, sh: "r", fcol: "#ffffff", scol: "#000000", swidth: 1.5, h: 20, w: 30},
					    						{x:  15, y: 0, sh: "c", fcol: "#ffffff", scol: "#000000", swidth: 1.5, sz: 20},
												{x: -2.5, y: 0, sh: "p", d: "M0 -10 L0 10", scol: "#000000", swidth: 1.5, sDash: [4]},
												{x: -10, y: 0, sh: "p", d: "M0 -10 L0 10", scol: "#000000", swidth: 1.5, sDash: [4]},
												{x: -17.5, y: 0, sh: "p", d: "M0 -10 L0 10", scol: "#000000", swidth: 1.5, sDash: [4]},
												{x:6.5, y:5, sh: "t", hbh: 5.5, hbw: 12,				   fcol: "#000000", sz: 12, txt:"PS", font:"Arial", w:0, h:0}
    										   ] },
		ft: "processorSharingService", tt: "Processor Sharing Queuing Station", l: "PS Server", p: [
				pcMap.D,
				pcMap.NumberServers,
				pcMap.Rt
		]
											    },
    {t: "n", d: {sh: "g", hbh: 20, hbw: 20,	d: [
					    						{x: 0, y: 0, sh: "c", fcol: "#777777", scol: "#000000", swidth: 1.5, sz: 40},
					    						{x: 0, y: 0, sh: "c", scol: "#ffffff", fcol: "#000000", swidth: 1.5, sz: 20}
    										   ] },
		ft: "sink", tt: "Sink (Open Models)", l: "Sink", p: [
		]
											    },
    {t: "n", d: {sh: "g", hbh: 20, hbw: 20,	d: [
					    						{x:  0, y: 10, sh: "e", fcol: "#ffffff", scol: "#000000", swidth: 1.5, h: 20, w:40},
					    						{x:  0, y: 0, sh: "r", fcol: "#ffffff", scol: "#ffffff", swidth: 1.5, h: 20, w:40},
					    						{x:  0, y: -10, sh: "e", fcol: "#ffffff", scol: "#000000", swidth: 1.5, h: 20, w:40},
												{x: -20, y: 0, sh: "p", d: "M0 -10 L0 11", scol: "#000000", swidth: 1.5},
												{x: -15, y: -5, sh: "p", d: "M0 0 L30 -10", scol: "#000000", swidth: 1.5},
												{x: -15, y: -15, sh: "p", d: "M0 0 L30 10", scol: "#000000", swidth: 1.5},
												{x: 20, y: 0, sh: "p", d: "M0 -10 L0 11", scol: "#000000", swidth: 1.5}
											   ] },
		ft: "router", tt: "Router", l: "Router", p: [
				pcMap.Rt
		]
											   },
    {t: "n", d: {sh: "g", hbh: 20, hbw: 20,	d: [
												{x:-8, y:10, sh: "t", hbh: 5.5, hbw: 12,		 fcol: "#0000AA", sz: 18, txt:"IN", font:"Arial", w:0, h:0},
												{x: 0, y: 0, sh: "p", d: "M20 5 L20 20 L-20 20 L-20 5 M-20 -5 L20 -20 L20 -5", scol: "#000000", swidth: 1.5}
    										   ] },
		ft: "referenceIn", tt: "System Entrance (Closed Models)", l: "System IN", p: [
				pcMap.Rt
		]
											    },
    {t: "n", d: {sh: "g", hbh: 20, hbw: 20,	d: [
												{x:-18.5, y:10, sh: "t", hbh: 5.5, hbw: 12,		 fcol: "#8800AA", sz: 18, txt:"OUT", font:"Arial", w:0, h:0},
												{x: 0, y: 0, sh: "p", d: "M20 5 L20 20 L-20 20 L-20 5 M-20 -5 L-20 -20 L20 -5", scol: "#000000", swidth: 1.5}
    										   ] },
		ft: "referenceOut", tt: "System Exit (Closed Models)", l: "System OUT", p: [
				pcMap.Rt
		]
											    },
    {t: "e", d: {lNorm: [{scol: "#000000", swidth: 1}],
    			 basArr:  {sh: "p", d: "M-5 1L0-2L5 2", scol: "#000000", swidth: 1},
    			 baeArr:  {sh: "p", d: "M-5-1L0 2L5-2", scol: "#000000", swidth: 1},
    			 eArr:  {sh: "p", d: "M-5-4L0 8L5-4z", fcol: "#000000", swidth: 1} },
		ft: "arc", tt: "Connection Arc", l: "Arc", p: [
				pcMap.Prob
				]}
];