"use strict"	// temporary - for a better debugging

// this function encapsulates the GUI
class DN4_FSM {
	constructor(tool) {
	    // all the "closure" variables
		this.DN4 = tool;			// the variable to access all the other components of the gui
	
	
		this.stateStack = false;	// meta keys handler can use this to return
									// to the states before the meta key was pressed
		
		// States of the GUI
		// base class
		class GUIstate {
			constructor(stId, tool) {
				this.Id = stId;
				this.DN4 = tool;
			}
			// Called when mouse moves in this state
			mousemove(event) {
				return this;
			}
			// Called when mouse is released in this state
			mouseup(event) {
				return this;
			}
			// Called when mouse is pressed in this state
			mousedown(event) {
				return this;
			}
			// Called when a timeout occurs in this state
			timeout() {
				return this;
			}
			// Called when the system enters in this state
			enter() {
			}
			// Called when the system exits this state
			exit() {
			}
		}
			
		// state specifications
		
		//	-- GS_NORMAL --
		class GS_NORMAL_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
			mousedown(event) {
				var qFSM = this.DN4.FSM;	// speed up access to FSM states and methods
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
		
				var metaHnd = qFSM.handleMeta(event);
		
				if(!metaHnd) {		// no meta-key combination pressed:
					var canX = (draw.xcorner + event.pageX) / draw.zoom;
					var canY = (draw.ycorner + event.pageY) / draw.zoom;
					if(draw.manipulatorAtPoint(canX, canY, ExtraGraph.shift(event))) {	// gives priority to manipulators
						return qFSM.GS_MANIPULATING;
					} else {										// otherwise selects the objects
						var underMouse = draw.atPoint(canX, canY);
						if(underMouse.length > 0) {	// if there is something under the mouse
							draw.underSelection = underMouse;	// sets objects that can be selected
							return qFSM.GS_SELECTING;		// go to the selection state
						} else {					// if there are no objects under the mouse
							// go to the selection rectangle state
							draw.selboxX = event.pageX;
							draw.selboxY = event.pageY - this.DN4.config.mainMenuSize;
							draw.selboxW = 0;
							draw.selboxH = 0;
							this.DN4.draw.toolTipText = "Drag: Rect. Selection";
							draw.drawView();

							return qFSM.GS_RECTSELECT;
						}
					}
					return qFSM.GS_NORMAL;			
				} else { // meta-key handling
					return metaHnd;
				}
			}
			mousemove(event) {
			  if(this.DN4.config.showMouseHover) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;

				var toUpdateDraw = false;

				if(draw.hoveringManipulatorAtPoint(canX, canY)) {	// gives priority to manipulators
					if(draw.prevHovering) {			// but there was an element under the mouse
						draw.prevHovering.s.hovering = false;
						draw.prevHovering = false;
						this.DN4.draw.toolTipText = "";
						toUpdateDraw = true;
						// updates the view
					}
					if(draw.prevHoveringManipNum != draw.hoveringManipNum) {	// over a new manipulator
					   	draw.hoveringManipX = draw.hoveringManipX * draw.zoom - draw.xcorner;
					   	draw.hoveringManipY = draw.hoveringManipY * draw.zoom - draw.ycorner;
//console.log(elementTypes[draw.hoveringManipEl.t].t + " " + draw.hoveringManipNum);
//						this.DN4.draw.toolTipText = "Manipulator " + elementTypes[draw.hoveringManipEl.t].t + " " + draw.hoveringManipNum;
//console.log(draw.hoveringManipEl);
//						this.DN4.draw.toolTipText = this.DN4.metaTypeElements[elementTypes[draw.hoveringManipEl.t].t].getManipulatorTooltip(draw.hoveringManipEl, draw.hoveringManipNum);
//console.log(draw.hoveringManip.handlers);
						this.DN4.draw.toolTipText = draw.hoveringManip.handlers.getManipulatorTooltip(draw.hoveringManipEl, draw.hoveringManipNum);
						draw.prevHoveringManipNum = draw.hoveringManipNum;
						toUpdateDraw = true;
					}
				} else {										// otherwise looks if it is over the objects
					if(draw.prevHoveringManipNum >= 0) {	// was hovering on a manipulator
						draw.prevHoveringManipNum = -1;		// clears it
						draw.hoveringManipNum = -1;
						this.DN4.draw.toolTipText = "";
						toUpdateDraw = true;
					}
					var underMouse = draw.atPoint(canX, canY);
					if(underMouse.length > 0) {	// if there is something under the mouse
						if(draw.prevHovering != underMouse[0]) { // if the mouse is over a new element
							underMouse[0].s.hovering = true;	// sets the element as "hovering"
							if(draw.prevHovering) {			// if there was an element under the mouse
								draw.prevHovering.s.hovering = false;
							}
							draw.prevHovering = underMouse[0];
							this.DN4.draw.toolTipText = "LMB: Select, SHIFT + LMB: Add to Selection";
							// updates the view
							toUpdateDraw = true;
						}
					} else { // nothing under the mouse now
						if(draw.prevHovering) {			// but there was an element unde the mouse
							draw.prevHovering.s.hovering = false;
							draw.prevHovering = false;
							// updates the view
							this.DN4.draw.toolTipText = "";
							toUpdateDraw = true;
						}
					}
				}
				if(toUpdateDraw)
				{
					draw.drawView();
				}
			  }
			  // hovering never changes state
			  return this.DN4.FSM.GS_NORMAL;
			}
			enter() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				if(draw.newElementType != -1) {	// clears any extra visualization needed to add elements
					draw.newElementType = -1;
					draw.drawView();
				}
			}
		}
		this.GS_NORMAL = new GS_NORMAL_GUIstate("normal", tool);
		
		//	-- GS_TRANSLATING --
		class GS_TRANSLATING_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
			mouseup(event) {
				return this.DN4.FSM.getPrevState();
			}
			mousemove(event) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				
				// updates the page offset
				draw.xcorner = draw.xcornerdown - event.pageX + draw.xmousedown;
				draw.ycorner = draw.ycornerdown - event.pageY + draw.ymousedown;
		
				// updates the view
				draw.drawView();
				
				return this.DN4.FSM.GS_TRANSLATING;
		    }
		}
		this.GS_TRANSLATING = new GS_TRANSLATING_GUIstate("translating", tool);
		
		//	-- GS_SCALING --
		class GS_SCALING_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
			mouseup(event) {
				return this.DN4.FSM.getPrevState();
			}
			mousemove(event) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				
				// updates the page offset
				draw.zoom = draw.zoomdown * Math.pow(2,
												(4*(event.pageX - draw.xmousedown) / draw.width)) ;
				draw.zoom = draw.zoom < this.DN4.config.zoomMin ? this.DN4.config.zoomMin :
									(draw.zoom > this.DN4.config.zoomMax ? this.DN4.config.zoomMax : draw.zoom);
		
		//		(xo + wm) / zo = (xn + wm) / zn  --->  xn = (xo + wm) / zo * zn - wm
				draw.xcorner = (draw.xcornerdown + draw.xmousedown) / draw.zoomdown * draw.zoom - draw.xmousedown;
				draw.ycorner = (draw.ycornerdown + draw.ymousedown) / draw.zoomdown * draw.zoom - draw.ymousedown;
				
				// updates the view
				draw.drawView();
		
				return this.DN4.FSM.GS_SCALING;
		    }
		}
		this.GS_SCALING = new GS_SCALING_GUIstate("scaling", tool);
	    
	    // -- GS_SELECTING --
		class GS_SELECTING_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
		    mouseup(event) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				
				var gmp = {cmd: "g"};	// to resotre group manipulators positions
				draw.saveGroupManipulators(gmp);
		
				var selel = draw.underSelection[0];
				if(selel.s.selected) {
					// the target element is selected
					if(ExtraGraph.shift(event))	{ // removes element from selection
		//				draw.removeFromCurrentSelection(selel);
						this.DN4.cmdHistory.addAndExec("Deselect",
									[{cmd: "s", add: [selel]}, gmp],	// undo
									[{cmd: "s", rem: [selel]}]		// command
							);
					} else {
						// does nothing! 
					}
				} else {
					// the target element is not selected
					if(!ExtraGraph.shift(event)) {	// clear the selection first if shift is not pressed
											// (change selection)
						this.DN4.cmdHistory.addAndExec("Select",
									[{cmd: "s", rem: "all"}, 
									 {cmd: "s", add: draw.currentSelectionToArray()}, gmp],	// undo
									[{cmd: "s", rem: "all"}, {cmd: "s", add: [selel]}]		// command
							);
					} else {	// adds to selection
						this.DN4.cmdHistory.addAndExec("Select",
									[{cmd: "s", rem: [selel]}, gmp],	// undo
									[{cmd: "s", add: [selel]}]		// command
							);
					}
			
			    	draw.underSelection = new Array();	// clear the possible selection
				}    	
				return this.DN4.FSM.GS_NORMAL;
		    }
		}
		this.GS_SELECTING = new GS_SELECTING_GUIstate("selecting", tool);
	
	    // -- GS_RECTSELECT --
		class GS_RECTSELECT_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
		    mouseup(event) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				var i;
				var undObj, redObj, p;
		
				var gmp = {cmd: "g"};	// to resotre group manipulators positions
				draw.saveGroupManipulators(gmp);
		
				undObj = [{cmd: "s", rem: "all"}, 
						  {cmd: "s", add: draw.currentSelectionToArray()}, gmp];
		
				// if a rectangle is made, but without shift key, clear the selection:
				// this selects only the newly selected element
				if(!ExtraGraph.shift(event)) {
					redObj = [{cmd: "s", rem: "all"}, {cmd: "s", add: []}];
					p = 1;
				} else {
					redObj = [{cmd: "s", add: []}];
					p = 0;
				}
				
				// clears the rect box
			  	if((draw.selboxW != 0) && (draw.selboxH != 0)) {
			  		var x, y, w, h, i, sel;
			  		
			  		x = draw.selboxW > 0 ? draw.selboxX : draw.selboxX + draw.selboxW;
			  		y = draw.selboxH > 0 ? draw.selboxY : draw.selboxY + draw.selboxH;
					draw.context.putImageData(draw.selboxImg, x - 1, y - 1);
					
					// finds the element selected
					x = (draw.xcorner + x) / draw.zoom;
					y = (draw.ycorner + y) / draw.zoom;
					w = (draw.selboxW > 0 ? draw.selboxW : -draw.selboxW) / draw.zoom;
					h = (draw.selboxH > 0 ? draw.selboxH : -draw.selboxH) / draw.zoom;
				  	sel = draw.selBox(x, y + this.DN4.config.mainMenuSize / draw.zoom, w, h);
				  	
				  	// adds the previously unselected elements to the current selection
				  	for(i = 0; i < sel.length; i++) {
				  		redObj[p].add[i] = sel[i];
				  	} 	
			  	}
			  	
			  	if((redObj[p].add.length > 0) ||
			  	   ( (draw.currentSelectionSize() > 0) && (!ExtraGraph.shift(event)) )) {
			  		// considers only usefull selections
					this.DN4.cmdHistory.addAndExec("Select", undObj, redObj);
			  	}
				this.DN4.draw.toolTipText = "";
				draw.drawView();
				return this.DN4.FSM.GS_NORMAL;
		    }
		    mousemove(event) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				
				// clears the rect box
			  	if((draw.selboxW != 0) && (draw.selboxH != 0)) {
					 draw.context.putImageData(draw.selboxImg,
					 			(draw.selboxW > 0 ? draw.selboxX : draw.selboxX + draw.selboxW) - 1,
					 			(draw.selboxH > 0 ? draw.selboxY : draw.selboxY + draw.selboxH) - 1);
			  	}
			  	draw.selboxW = event.pageX - draw.selboxX;
			  	draw.selboxH = event.pageY - this.DN4.config.mainMenuSize - draw.selboxY;
			  	// draw the new rectbox
			  	if((draw.selboxW != 0) && (draw.selboxH != 0)) {
			  		// gets the background
					draw.selboxImg = draw.context.getImageData(
					 			(draw.selboxW > 0 ? draw.selboxX : draw.selboxX + draw.selboxW)-1,
					 			(draw.selboxH > 0 ? draw.selboxY : draw.selboxY + draw.selboxH)-1,
					 			(draw.selboxW > 0 ? draw.selboxW : -draw.selboxW) + 2,
					 			(draw.selboxH > 0 ? draw.selboxH : -draw.selboxH) + 2);
					// draws the selection box
				  	draw.context.strokeStyle = "#000000";		// sets the color
				  	draw.context.strokeRect(draw.selboxX, draw.selboxY,
				  						    draw.selboxW, draw.selboxH);
			  	}
				return this.DN4.FSM.GS_RECTSELECT;
		    }
		}
		this.GS_RECTSELECT = new GS_RECTSELECT_GUIstate("rectselect", tool);
	
	    // -- GS_MANIPULATING --
		class GS_MANIPULATING_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
			mouseup(event) {
				// calls the appropriate manipulation function
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
			   	var cm = this.DN4.draw.underManipulation;
			   	cm.handlers.end(canX, canY, ExtraGraph.shift(event), cm);
		
				if((event.which == 3) && (!draw.dontShowAdvancedManipulationMenu)) {
					// if the right mouse button can open the advanced menu
					var newState = cm.handlers.enterAdvanced(cm);
					if(newState) {
						if(newState === this.DN4.FSM.GS_ADVMENUSHOWN) {
							var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
							draw.enteringAdvancedManipulationMenu = false;
						}
						draw.drawView();
					   	return newState;
					}
				}
				
				draw.drawView();
				return this.DN4.FSM.GS_NORMAL;
			}
			mousemove(event) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				// stops the timer to show advanced manipulators
				ExtraGraph.stopTimer();
				// if right button was clicked, disable the show of menu
				if(event.which == 3) {
					draw.dontShowAdvancedManipulationMenu = true;
				}
				// calls the appropriate manipulation function
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
			   	var cm = this.DN4.draw.underManipulation;
			   	cm.handlers.move(canX, canY, ExtraGraph.shift(event), cm);
		
				draw.drawView();	
		
				return this.DN4.FSM.GS_MANIPULATING;
			}
			timeout() {
			   	var cm = this.DN4.draw.underManipulation;
				var newState = cm.handlers.enterAdvanced(cm);
				if(newState) {
					if(newState === this.DN4.FSM.GS_ADVMENUSHOWN) {
						var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
						draw.enteringAdvancedManipulationMenu = true;
					}
				   	this.DN4.draw.drawView();
				   	return newState;
				}
			   	return this.DN4.FSM.GS_MANIPULATING;
			}
			enter() {
				ExtraGraph.startTimer(this.DN4.config.objMenuClickTimeout);
				this.DN4.draw.dontShowAdvancedManipulationMenu = false;	// to handle right click menu open
			}
			exit() {
				ExtraGraph.stopTimer();
			}
		}
		this.GS_MANIPULATING = new GS_MANIPULATING_GUIstate("manipulating", tool);
	
		//	-- GS_ADVMENUSHOWN --
		class GS_ADVMENUSHOWN_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
			mouseup(event) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
				var i;
		
				// if inside the menu
				if((canX >= draw.advancedManipulatorMenu.x) &&
				   (canX <= draw.advancedManipulatorMenu.x +		
				   			this.DN4.config.textBaseSize * draw.advancedManipulatorMenu.w)) {
				   	i = Math.floor((canY - draw.advancedManipulatorMenu.y) /
				   	       (this.DN4.config.textBaseSize * 1.5)+1);
				   	if((i >= 0) && (i < draw.advancedManipulatorMenu.items.length)) {
				   		var newState = draw.advancedManipulatorMenu.items[i].act();
						draw.advancedManipulatorMenu = false;
						draw.advancedManipulatroMenuSelected = -1;
						draw.drawView();
						if(newState) {
							return newState;
						} else {
							return this.DN4.FSM.GS_NORMAL;
						}
				   	}
				}
				if(draw.enteringAdvancedManipulationMenu) {
					draw.enteringAdvancedManipulationMenu = false;
					return this.DN4.FSM.GS_ADVMENUSHOWN;
				} else {
					draw.advancedManipulatorMenu = false;
					draw.advancedManipulatroMenuSelected = -1;
					draw.drawView();
					return this.DN4.FSM.GS_NORMAL;
				}
			}
			mousemove(event) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
				var i;
				
				// if inside the menu
				if((canX >= draw.advancedManipulatorMenu.x) &&
				   (canX <= draw.advancedManipulatorMenu.x +		
				   			this.DN4.config.textBaseSize * draw.advancedManipulatorMenu.w)) {
				   	i = Math.floor((canY - draw.advancedManipulatorMenu.y) /
				   	       (this.DN4.config.textBaseSize * 1.5)+1);
				   	if((i >= 0) && (i < draw.advancedManipulatorMenu.items.length)) {
				   		if(draw.advancedManipulatroMenuSelected != i) {
					   		draw.advancedManipulatroMenuSelected = i;
							draw.drawView();
				   		}
				   	} else {
				   		if(draw.advancedManipulatroMenuSelected != -1) {
					   		draw.advancedManipulatroMenuSelected = -1;
							draw.drawView();
				   		}
				   	}
				}
			
				return this.DN4.FSM.GS_ADVMENUSHOWN;
			}
		}
		this.GS_ADVMENUSHOWN = new GS_ADVMENUSHOWN_GUIstate("advmenushown", tool);
		
		//	-- GS_ADDINTPOINT --
		// adds an intermediate point to an edge
		class GS_ADDINTPOINT_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
			mousedown(event) {
				var qFSM = this.DN4.FSM;	// speed up access to FSM states and methods
				var metaHnd = qFSM.handleMeta(event);
		
				if(!metaHnd) {		// no meta-key combination pressed:
					return qFSM.GS_ADDINTPOINT;
				} else { // meta-key handling
					return metaHnd;
				}
			}
			mouseup(event) {
				var qFSM = this.DN4.FSM;	// speed up access to FSM states and methods
		
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
				var underMouse = draw.atPoint(canX, canY);
				if(underMouse.length > 0) {
					var et = elementTypes[underMouse[0].t];		// get the element type
					
					if(et.t == 'e') { // intermediate point can be added only to edges
						this.DN4.metaTypeElements['e'].addIntermediatePoint(
														underMouse[0],
														underMouse[0].s, et.d,
														canX, canY);
					}
				}	
				// disable add point hovering
				draw.addPointHovering  = false;
				return qFSM.GS_NORMAL;
			}
			mousemove(event) {
			  if(this.DN4.config.showMouseHover) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
				var underMouse = draw.atPoint(canX, canY);
				if(underMouse.length > 0) {	// if there is something under the mouse
				  if(underMouse[0].s.selected) { // considers only selected elements
					if(draw.prevHovering != underMouse[0]) { // if the mouse is over a new element
						underMouse[0].s.hovering = true;	// sets the element as "hovering"
						if(draw.prevHovering) {			// if there was an element under the mouse
							draw.prevHovering.s.hovering = false;
						}
						draw.prevHovering = underMouse[0];
						
						// adds the new intermeidate point marker
						draw.addPointHovering  = true;
						draw.addPointHoveringX = canX;
						draw.addPointHoveringY = canY;
						// updates the view
						draw.drawView();
					} else {
					  if(draw.addPointHovering) {
						draw.addPointHoveringX = canX;
						draw.addPointHoveringY = canY;
						// updates the view
						draw.drawView();
					  }
					}
				  }
				} else { // nothing under the mouse now
					if(draw.prevHovering) {			// but there was an element unde the mouse
						draw.prevHovering.s.hovering = false;
						draw.prevHovering = false;
						// removes the new intermeidate point marker
						draw.addPointHovering  = false;
						// updates the view
						draw.drawView();
					}
				}
			  }
			  // hovering never changes state
			  return this.DN4.FSM.GS_ADDINTPOINT;
			}
			enter() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				draw.currentModeText = "Add curve point";
				draw.drawView();
			}
			exit() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				draw.currentModeText = "";
				draw.drawView();
			}
		}
		this.GS_ADDINTPOINT = new GS_ADDINTPOINT_GUIstate("addintpoint", tool);
	
		//	-- GS_ADDNODE --
		// adds a node to the model
		class GS_ADDNODE_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
			mousedown(event) {
				var qFSM = this.DN4.FSM;	// speed up access to FSM states and methods
				var metaHnd = qFSM.handleMeta(event);
		
				if(!metaHnd) {		// no meta-key combination pressed:
					return qFSM.GS_ADDNODE;
				} else { // meta-key handling
					return metaHnd;
				}
			}
			mouseup(event) {
				var qFSM  = this.DN4.FSM;	// speed up access to FSM states and methods
				var draw  = this.DN4.draw;	// speed up accessing the draw object of the GUI
				var model = this.DN4.model;	// speed up accessing the model methods
				var ctrl  = this.DN4.controls;	// speed up accessing the control methods
//// SNAP SUPPORT ?
//				var canX = (draw.xcorner + event.pageX) / draw.zoom;
//				var canY = (draw.ycorner + event.pageY) / draw.zoom;
				var canX = ExtraGraph.snap((draw.xcorner + event.pageX) / draw.zoom);
				var canY = ExtraGraph.snap((draw.ycorner + event.pageY) / draw.zoom);
//// SNAP SUPPORT ? (END)
				var underMouse = draw.atPoint(canX, canY);
				if(underMouse.length == 0) {
					// adding a node
					var el = model.getFree();
					if(ctrl.addingMode == 1) { // single element addition: selects it and return to normal mode
						var gmp = {cmd: "g"};	// to resotre group manipulators positions
						draw.saveGroupManipulators(gmp);
						this.DN4.cmdHistory.addAndExec("Add Node",
							[{cmd: "s", rem: "all"}, 					// UNDO (with re-selection) 
							 {cmd: "s", add: draw.currentSelectionToArray()}, gmp,
							 {cmd: "m", rem: [el]}],
							[{cmd: "m", add: [{el: el, t: ctrl.selectedEl,   // command
											   s: {tx: canX, ty: canY,
											       sx: 1.0, sy:1, rot: 0, in: [], out: []}
											   }]
		/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/
									  , ff: graphFirstFree
		/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/
							 },
							 {cmd: "s", rem: "all"}, {cmd: "s", add: [el]}]
						);
						draw.drawView();
						ctrl.selectedEl = -1;
						ctrl.draw();
						return qFSM.GS_NORMAL;
					} else {	// multiple selection mode
						this.DN4.cmdHistory.addAndExec("Add Node",
							[{cmd: "m", rem: [el]}], 					// UNDO
							[{cmd: "m", add: [{el: el, t: ctrl.selectedEl,
											   s: {tx: canX, ty: canY,
											       sx: 1.0, sy:1, rot: 0, in: [], out: []}
											   }]
		/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/
									  , ff: graphFirstFree
		/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/
							 }] // REDO
						);
						draw.drawView();
						return qFSM.GS_ADDNODE;
					}
				} else {	// if the element cannot be added here, mouseup does nothing
					return qFSM.GS_ADDNODE;
				}
			}
			mousemove(event) {
			  if(this.DN4.config.showMouseHover) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				
//// SNAP SUPPORT ?
//				var canX = (draw.xcorner + event.pageX) / draw.zoom;
//				var canY = (draw.ycorner + event.pageY) / draw.zoom;
				var canX = ExtraGraph.snap((draw.xcorner + event.pageX) / draw.zoom);
				var canY = ExtraGraph.snap((draw.ycorner + event.pageY) / draw.zoom);
//// SNAP SUPPORT ? (END)
				var underMouse = draw.atPoint(canX, canY);
				draw.newNodeX = canX;
				draw.newNodeY = canY;
				if(underMouse.length == 0) {	// if there is nothing under the mouse
					draw.newElementType = this.DN4.controls.selectedEl;
				} else {
					draw.newElementType = -2;		// uses -2 to show a cross when a node cannto be added
				}
				// updates the view
				draw.drawView();
			  }
			  return this.DN4.FSM.GS_ADDNODE;
			}
			enter() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				draw.currentModeText = "Add element (node)";
				draw.newElementType = -1;
				draw.drawView();
			}
			exit() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				draw.currentModeText = "";
				draw.drawView();
			}
		}
		this.GS_ADDNODE = new GS_ADDNODE_GUIstate("addnode", tool);
	
		//	-- GS_ADDEDGE --
		// adds an edge to the model
		class GS_ADDEDGE_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
			mousedown(event) {
				var qFSM = this.DN4.FSM;	// speed up access to FSM states and methods
				var metaHnd = qFSM.handleMeta(event);
		
				if(!metaHnd) {		// no meta-key combination pressed:
					var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
					var canX = (draw.xcorner + event.pageX) / draw.zoom;
					var canY = (draw.ycorner + event.pageY) / draw.zoom;
					var underMouse = draw.atPoint(canX, canY);
					var tt;
					
					if(underMouse.length > 0) {	// if there is something under the mouse
					  tt = elementTypes[underMouse[0].t].t;
					} else {
					  tt = '-';
					}
					if(tt == 'n') {  // for now, edges can only start from nodes
						// starts editing the tangent of the new arc
						draw.newEdgeXc1 = draw.newEdgeX1 = underMouse[0].s.tx;
						draw.newEdgeYc1 = draw.newEdgeY1 = underMouse[0].s.ty;
						draw.newEdgeFrom = underMouse[0].id;
						draw.newElementType = this.DN4.controls.selectedEl;
						return qFSM.GS_ADDEDGE1STTG;
					}	
				    return this.DN4.FSM.GS_ADDEDGE;
				} else { // meta-key handling
					return metaHnd;
				}
			}
			mousemove(event) {
			  if(this.DN4.config.showMouseHover) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
				var underMouse = draw.atPoint(canX, canY);
				var tt;
				
				if(underMouse.length > 0) {	// if there is something under the mouse
				  tt = elementTypes[underMouse[0].t].t;
				} else {
				  tt = '-';
				}
				if(tt == 'n') {  // for now, edges can only start from nodes
		
					if(draw.prevHovering != underMouse[0]) { // if the mouse is over a new element
						underMouse[0].s.hovering = true;	// sets the element as "hovering"
						if(draw.prevHovering) {			// if there was an element under the mouse
							draw.prevHovering.s.hovering = false;
						}
						draw.prevHovering = underMouse[0];
						draw.newElementType = -1;
						// updates the view
						draw.drawView();
					}
				} else { // nothing under the mouse now
					if(draw.prevHovering) {			// but there was an element unde the mouse
						draw.prevHovering.s.hovering = false;
						draw.prevHovering = false;
					}
					draw.newNodeX = canX;
					draw.newNodeY = canY;
					draw.newElementType = -2;		// uses -2 to show a cross when a node cannto be added
					// updates the view
					draw.drawView();
				}
			  }
			  return this.DN4.FSM.GS_ADDEDGE;
			}
			enter() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				draw.currentModeText = "Add element (edge)";
				draw.newElementType = -1;
				draw.newEdgeFrom = -1;
				draw.newEdgeTo   = -1;
				draw.newEdgePoints = [];
				draw.newEdge1stTangent = false;
				draw.drawView();
			}
			exit() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				draw.currentModeText = "";
				draw.drawView();
			}
		}
		this.GS_ADDEDGE = new GS_ADDEDGE_GUIstate("addedge", tool);
	
		//	-- GS_ADDEDGE1STTG --
		// adds an edge to the model - sets the first tangent
		class GS_ADDEDGE1STTG_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
			mouseup(event) {
				var qFSM = this.DN4.FSM;	// speed up access to FSM states and methods
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
				if((canX - draw.newEdgeX1) * (canX - draw.newEdgeX1) +
				   (canY - draw.newEdgeY1) * (canY - draw.newEdgeY1) > 36) {	// needs a minimum dispalcement
					draw.newEdgeXc1 = canX;										// to set the tangent
					draw.newEdgeYc1 = canY;
				}
				return qFSM.GS_ADDEDGEINT;
			}
			mousemove(event) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
//// SNAP SUPPORT ?
				var canXQ = ExtraGraph.snap(canX);
				var canYQ = ExtraGraph.snap(canY);
//// SNAP SUPPORT ? (END)
				if((canX - draw.newEdgeX1) * (canX - draw.newEdgeX1) +
				   (canY - draw.newEdgeY1) * (canY - draw.newEdgeY1) > 36) {	// needs a minimum dispalcement
					draw.newEdgeXc1 = canXQ;										// to set the tangent
					draw.newEdgeYc1 = canYQ;
					draw.drawView();
				}
			    return this.DN4.FSM.GS_ADDEDGE1STTG;
			}
			enter() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				draw.currentModeText = "Add element (edge) 1st tangent";
				draw.newEdge1stTangent = true;
				draw.drawView();
			}
			exit() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				draw.currentModeText = "";
				draw.newEdge1stTangent = false;
				draw.drawView();
			}
		}
		this.GS_ADDEDGE1STTG = new GS_ADDEDGE1STTG_GUIstate("addedge1sttg", tool);
		
		//	-- GS_ADDEDGEINT --
		// adds an edge to the model - sets the first tangent
		class GS_ADDEDGEINT_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
			mousedown(event) {
				var qFSM = this.DN4.FSM;	// speed up access to FSM states and methods
				var metaHnd = qFSM.handleMeta(event);
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
		
				if(!metaHnd) {		// no meta-key combination pressed:
					var canX = (draw.xcorner + event.pageX) / draw.zoom;
					var canY = (draw.ycorner + event.pageY) / draw.zoom;
//// SNAP SUPPORT ?
					var canXQ = ExtraGraph.snap(canX);
					var canYQ = ExtraGraph.snap(canY);
//// SNAP SUPPORT ? (END)
					var underMouse = draw.atPoint(canX, canY);
					var tt;
					
					draw.newEdgeXc2 = draw.newEdgeX2 = canXQ;
					draw.newEdgeYc2 = draw.newEdgeY2 = canYQ;
		
					if(underMouse.length > 0) {	// if there is something under the mouse
					  tt = elementTypes[underMouse[0].t].t;
					} else {
					  tt = '-';
					}
					if(tt == 'n') {  // for now, edges can only start from nodes
						// starts editing the tangent of the new arc
						draw.newEdgeTo = underMouse[0].id;
						draw.newEdgeXc2 = draw.newEdgeX2 = underMouse[0].s.tx;
						draw.newEdgeYc2 = draw.newEdgeY2 = underMouse[0].s.ty;
						return qFSM.GS_ADDEDGE2NDTG;
					} else if(tt == '-') {
						return qFSM.GS_ADDEDGE2NDTG;
					} else {	// Trying to end on an edge
					    return this.DN4.FSM.GS_ADDEDGEINT;
					}
				} else { // meta-key handling
					draw.fromMeta = true;
					return metaHnd;
				}
			}
			mousemove(event) {
			  if(this.DN4.config.showMouseHover) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
//// SNAP SUPPORT ?
				var canXQ = ExtraGraph.snap(canX);
				var canYQ = ExtraGraph.snap(canY);
//// SNAP SUPPORT ? (END)
				var underMouse = draw.atPoint(canX, canY);
				var tt;
				
				draw.newEdgeXc2 = draw.newEdgeX2 = canXQ;
				draw.newEdgeYc2 = draw.newEdgeY2 = canYQ;
		
				if(underMouse.length > 0) {	// if there is something under the mouse
				  tt = elementTypes[underMouse[0].t].t;
				} else {
				  tt = '-';
				}
				if(tt == 'n') {  // for now, edges can only start from nodes
		
					if(draw.prevHovering != underMouse[0]) { // if the mouse is over a new element
						underMouse[0].s.hovering = true;	// sets the element as "hovering"
						if(draw.prevHovering) {			// if there was an element under the mouse
							draw.prevHovering.s.hovering = false;
						}
						draw.prevHovering = underMouse[0];
						draw.newElementType = this.DN4.controls.selectedEl;
						// updates the view
						draw.drawView();
					}
				} else { // nothing under the mouse now
					if(draw.prevHovering) {			// but there was an element unde the mouse
						draw.prevHovering.s.hovering = false;
						draw.prevHovering = false;
					}
					if(tt == '-') {
						draw.newElementType = this.DN4.controls.selectedEl;
					} else {
						draw.newNodeX = canXQ;
						draw.newNodeY = canYQ;
						draw.newElementType = -2;		// uses -2 to show a cross when a node cannot be added
					}
					// updates the view
					draw.drawView();
				}
			  }
			  return this.DN4.FSM.GS_ADDEDGEINT;
			}
			enter() {
			  var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
			  draw.currentModeText = "Add element (edge) intermediate point";
			  if(draw.fromMeta) {
			  	draw.fromMeta = false;
			  } else {
				var i = draw.newEdgePoints.length;
				if(i > 0) {
					draw.newEdgeX1  = draw.newEdgePoints[i-1].x2;
					draw.newEdgeY1  = draw.newEdgePoints[i-1].y2;
					draw.newEdgeXc1 = 2 * draw.newEdgePoints[i-1].x2 - draw.newEdgePoints[i-1].xc2;
					draw.newEdgeYc1 = 2 * draw.newEdgePoints[i-1].y2 - draw.newEdgePoints[i-1].yc2;
				}
				draw.newEdgeXc2 = draw.newEdgeXc1;
				draw.newEdgeYc2 = draw.newEdgeYc1;
				draw.newEdgeX2  = draw.newEdgeXc1;
				draw.newEdgeY2  = draw.newEdgeYc1;
				draw.newEdgePoints[i] = {
					x1:  draw.newEdgeX1,  y1:  draw.newEdgeY1,  xc1: draw.newEdgeXc1, yc1: draw.newEdgeYc1,
					t1: 7,
					xc2: draw.newEdgeXc2, yc2: draw.newEdgeYc2, x2:  draw.newEdgeX2,  y2:  draw.newEdgeY2,
					t2: 7
				};
				draw.drawView();
			  }
			}
			exit() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				draw.currentModeText = "";
				draw.drawView();
			}
		}
		this.GS_ADDEDGEINT = new GS_ADDEDGEINT_GUIstate("addedgeint", tool);		
	
		//	-- GS_ADDEDGE2NDTG --
		// adds an edge to the model - sets the first tangent
		class GS_ADDEDGE2NDTG_GUIstate extends GUIstate {
			constructor(stId, tool) {
				super(stId, tool);
			}
			mouseup(event) {
				var qFSM = this.DN4.FSM;	// speed up access to FSM states and methods
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				var model = this.DN4.model;	// speed up accessing the model methods
				var ctrl = this.DN4.controls;	// speed up accessing the control methods
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
//// SNAP SUPPORT ?
				var canXQ = ExtraGraph.snap(canX);
				var canYQ = ExtraGraph.snap(canY);
//// SNAP SUPPORT ? (END)
				if((canX - draw.newEdgeX2) * (canX - draw.newEdgeX2) +
				   (canY - draw.newEdgeY2) * (canY - draw.newEdgeY2) > 36) {	// needs a minimum dispalcement
					draw.newEdgeXc2 = 2 * draw.newEdgeX2 - canXQ;				// to set the tangent
					draw.newEdgeYc2 = 2 * draw.newEdgeY2 - canYQ;
					draw.drawView();
				}
		
				var i = draw.newEdgePoints.length-1;
				draw.newEdgePoints[i] = {
					x1:  draw.newEdgeX1,  y1:  draw.newEdgeY1,  xc1: draw.newEdgeXc1, yc1: draw.newEdgeYc1,
					t1: 7,
					xc2: draw.newEdgeXc2, yc2: draw.newEdgeYc2, x2:  draw.newEdgeX2,  y2:  draw.newEdgeY2,
					t2: 7
				};
		
				if(draw.newEdgeTo == -1) {
					return qFSM.GS_ADDEDGEINT;
				} else {
					// adding a node
					var el = model.getFree();
		//alert("Adding edge <" + el.id + "> from <" + draw.newEdgeFrom + "> to <" + draw.newEdgeTo + ">");
					var s = {
						    	from:   draw.newEdgeFrom,
						    	to:     draw.newEdgeTo,
						    	points: draw.newEdgePoints
						    };
					var fromEl = graph[draw.newEdgeFrom];
					var fromOutPos = fromEl.s.out.length;
					var toEl = graph[draw.newEdgeTo];
					var toInPos = toEl.s.in.length;
					if(ctrl.addingMode == 1) { // single element addition: selects it and return to normal mode
						var gmp = {cmd: "g"};	// to resotre group manipulators positions
						draw.saveGroupManipulators(gmp);
						this.DN4.cmdHistory.addAndExec("Add Edge",
							[{cmd: "s", rem: "all"}, 					// UNDO (with re-selection) 
							 {cmd: "s", add: draw.currentSelectionToArray()}, gmp,
							 {cmd: "m", ref: [{el: el,     dr: -2},     	// adjusts the references
							 				  {el: fromEl, dr: -1},
							 				  {el: toEl,   dr: -1}]},
							 {cmd: "m", rem: [el]},	// removes the element
							 {cmd: "p", el: fromEl, path: ["out"], rem: fromOutPos}, // adjust from out array
							 {cmd: "p", el: toEl, path: ["in"], rem: toInPos} // adjust to in array
							],
							[{cmd: "m", add: [{el: el, t: ctrl.selectedEl,   // command
											   s: s
											   }]
		/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/
									  , ff: graphFirstFree
		/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/
											   },
							 {cmd: "m", ref: [{el: el,     dr: 2},     	// adjusts the references
							 				  {el: fromEl, dr: 1},
							 				  {el: toEl,   dr: 1}]},
							 {cmd: "p", el: fromEl, path: ["out"], add: fromOutPos}, // adjust from out array
							 {cmd: "p", el: fromEl, path: ["out", fromOutPos], val: el.id},
							 {cmd: "p", el: toEl, path: ["in"], add: toInPos}, // adjust to in array
							 {cmd: "p", el: toEl, path: ["in", toInPos], val: el.id}, // adjust to in array
							 {cmd: "s", rem: "all"}, {cmd: "s", add: [el]}]
						);
						ctrl.selectedEl = -1;
						draw.newElementType = -1;
						draw.newEdgeFrom = -1;
						draw.newEdgeTo   = -1;
						draw.newEdgePoints = [];
						ctrl.draw();
						return qFSM.GS_NORMAL;
					} else {	// multiple selection mode
						this.DN4.cmdHistory.addAndExec("Add Edge",
							[{cmd: "m", ref: [{el: el,     dr: -2},     	// adjusts the references
							 				  {el: fromEl, dr: -1},
							 				  {el: toEl,   dr: -1}]},
							 {cmd: "m", rem: [el]},
							 {cmd: "p", el: fromEl, path: ["out"], rem: fromOutPos}, // adjust from out array
							 {cmd: "p", el: toEl, path: ["in"], rem: toInPos} // adjust to in array
							], 					// UNDO
							[{cmd: "m", add: [{el: el, t: ctrl.selectedEl,
											   s: s
											   }]
		/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/
									  , ff: graphFirstFree
		/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/
							 },
							 {cmd: "m", ref: [{el: el,     dr: 2},     	// adjusts the references
							 				  {el: fromEl, dr: 1},
							 				  {el: toEl,   dr: 1}]},
							 {cmd: "p", el: fromEl, path: ["out"], add: fromOutPos}, // adjust from out array
							 {cmd: "p", el: fromEl, path: ["out", fromOutPos], val: el.id},
							 {cmd: "p", el: toEl, path: ["in"], add: toInPos}, // adjust to in array
							 {cmd: "p", el: toEl, path: ["in", toInPos], val: el.id} // adjust to in array
							] // command
						);
						return qFSM.GS_ADDEDGE;
					}
				}
			}
			mousemove(event) {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				var canX = (draw.xcorner + event.pageX) / draw.zoom;
				var canY = (draw.ycorner + event.pageY) / draw.zoom;
//// SNAP SUPPORT ?
				var canXQ = ExtraGraph.snap(canX);
				var canYQ = ExtraGraph.snap(canY);
//// SNAP SUPPORT ? (END)
				if((canX - draw.newEdgeX2) * (canX - draw.newEdgeX2) +
				   (canY - draw.newEdgeY2) * (canY - draw.newEdgeY2) > 36) {	// needs a minimum dispalcement
					draw.newEdgeXc2 = 2 * draw.newEdgeX2 - canXQ;				// to set the tangent
					draw.newEdgeYc2 = 2 * draw.newEdgeY2 - canYQ;
					draw.drawView();
				}
			    return this.DN4.FSM.GS_ADDEDGE2NDTG;
			}
			enter() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				draw.currentModeText = "Add element (edge) 2nd tangent";
				draw.newEdge2ndTangent = true;
				draw.drawView();
			}
			exit() {
				var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
				draw.currentModeText = "";
				draw.newEdge2ndTangent = false;
				draw.drawView();
			}
		}
		this.GS_ADDEDGE2NDTG = new GS_ADDEDGE2NDTG_GUIstate("addedge2ndtg", tool);

	}
	
	getPrevState() {	// returns the previous state before meta and resets it
		var s = this.stateStack;
		this.stateStack = false;
		return s;
	}
	// handles meta keys in a consistent way in all the FSM states
	handleMeta(event) {
		var metaCodes = (ExtraGraph.shift(event) ? 1 : 0) +
						(ExtraGraph.ctrl(event)  ? 2 : 0) +
						(ExtraGraph.alt(event)   ? 4 : 0);
		var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
		draw.xcornerdown = draw.xcorner;	// copies the position of the top-left corner
		draw.ycornerdown = draw.ycorner;
		draw.xmousedown = event.pageX;	// copies the position of the mouse
		draw.ymousedown = event.pageY;
		draw.zoomdown = draw.zoom;			// zoom when the mouse was pressed

		if(metaCodes <= 1) {		// SHIFT or no key pressed:
			return false;			
		} else if(metaCodes == 2) {	// with ctrl key only, scales the screen
			this.stateStack = this.DN4.guiState;
			return this.GS_SCALING;
		} else if(metaCodes == 4) {	// with alt  key only, translates the screen
			this.stateStack = this.DN4.guiState;
			return this.GS_TRANSLATING;
		}

		// unknown keyboard combo - treat as no meta
		return false;
	}

};