"use strict"	// temporary - for a better debugging

// this function encapsulates the class of drawing part of the GUI
class DN4_controls {
	constructor(tool) {
	    // all the "closure" variables
		this.DN4 = tool;			// the variable to access all the other components of the gui
	
		// html related
	    this.canvas = document.getElementById("ctrlCanvas");	// the canvas component
	    this.context = this.canvas.getContext("2d");			// the drawing context of the canvas
	    // height and width of the window
	    this.width  = this.canvas.width  = this.DN4.config.controlMenuSize;	
	    this.height = this.canvas.height = $(window).height() - this.DN4.config.mainMenuSize;
	    this.top    = this.DN4.config.mainMenuSize;
	    this.left   = $(window).width() - this.DN4.config.controlMenuSize;
	
		$("#ctrlCanvas").offset({ top: this.top, left: this.left})
	
		// selected element
		this.selectedEl = -1;
		// adding mode: 1 - add one element
		//              2 - add many elements
		this.addingMode = 1;
		// hovering element
		this.hoveringEl = -1;
		// scroll menu mode
		this.menuScrolling = false;
		// position where the mouse was pressed;
		this.xmousedown = 0;
		this.ymousedown = 0;
		// mouse button is currently pressed
		this.dragging = false;
		// to handle scrolling
		this.scrollmode = false;
		this.scrollY = 0;
		this.scrollYatdown = 0;
		
		// to see if the key was pressed for long enough
		this.timeDown = 0;	
	}	
	// Initialization
	init() {
		this.draw();
	};

	// drawing
	draw() {
	    var context = this.context; // quick access to the context
	    var i;
	    var et;	// element type of the current element
	    var cx = this.DN4.config.controlMenuSize / 2;
	    var cy = this.DN4.config.controlMenuSpace / 2 - this.scrollY;
	    var dx = this.DN4.config.controlMenuSpace;
	    var dy = this.DN4.config.controlMenuSpace;

		context.clearRect(0, 0, this.width, this.height);
		context.fillStyle="#ffffff";
		context.fillRect(0, 0, this.width, this.height);
		
		if(this.hoveringEl >= 0) {
		  	context.fillStyle = "#ff9999";		// sets the color
			context.fillRect(cx - dx * 0.5, cy + dy * (this.hoveringEl - 0.5), dx, dy);
		}

		for(i = 0; i < elementTypes.length; i++) {
			et = elementTypes[i];
			this.DN4.metaTypeElements[et.t].draw(context,
				et.t == 'n' ?
				    {
				    	tx: cx,
				    	ty: cy + dy * i,
				    	sx: 1.0, sy:1, rot: 0, in: [], out: []
				    } : {
				    	from: 0, to: 0,
				    	points: [{
				    		x1: cx - dx * 0.5,
				    		y1: cy + dy * (i + 0.5),
				    		xc1: cx - dx * 0.25,
				    		yc1: cy + dy * (i - 0.25),
				    		t1: 7,
				    		xc2: cx - dx * 0.5,
				    		yc2: cy + dy * (i - 0.125),
				    		x2: cx + dx * 0.5,
				    		y2: cy + dy * (i - 0.5),
				    		t2: 7
				    	}]
				    }, et.d);
		}
		
		if(this.selectedEl >= 0) {
			if(this.addingMode == 2) {
				context.lineWidth = 5;
			  	context.strokeStyle = "#008800";		// sets the color
			} else {
				context.lineWidth = 3;
			  	context.strokeStyle = "#ff0000";		// sets the color
			}
			context.strokeRect(cx - dx * 0.5, cy + dy * (this.selectedEl - 0.5), dx, dy);
		}
	};

	// resizes the window
	resize() {
	    this.height = this.canvas.height = $(window).height() - this.DN4.config.mainMenuSize;		
	    this.left   = $(window).width() - this.DN4.config.controlMenuSize;
		$("#ctrlCanvas").offset({ top: this.top, left: this.left});

		// checks if the scrolling must be updated
		if(this.scrollY > elementTypes.length * this.DN4.config.controlMenuSpace -
		                  this.height) {
		       this.scrollY = elementTypes.length * this.DN4.config.controlMenuSpace -
		                  this.height};
		if(this.scrollY < 0) {this.scrollY = 0};
	    this.draw();
	}	

	// handles mouseup
	mouseup(event) {
		var canX = event.pageX - this.left;
		var canY = event.pageY - this.top;

		if(!this.scrollmode) {	// scrolling disables element selections
			if(this.hoveringEl >= 0) {	// mouseup occurred over an eleemnt
				if(this.hoveringEl == this.selectedEl) {	// deselect, if selected
					this.selectedEl = -1;
					this.DN4.setState(this.DN4.FSM.GS_NORMAL);
				} else {	// otherwise, selects the element where the mouse is hovering
					this.selectedEl = this.hoveringEl;
					// checks the metatype of the element
					var met = elementTypes[this.selectedEl].t;
					switch(met) {
					  case 'n':
						this.DN4.setState(this.DN4.FSM.GS_ADDNODE);
						break;
					  case 'e':
						this.DN4.setState(this.DN4.FSM.GS_ADDEDGE);
						break;
					  default:
						alert("adding a currently unsupported meta element type");
						this.DN4.setState(this.DN4.FSM.GS_NORMAL);
						break;
					}
					// checks whether it has been left-clicked or right clicked
					var deltaT = (new Date).getTime() - this.timeDown;
					if((event.which == 3) || (deltaT > this.DN4.config.objMenuClickTimeout)) {
						// right clicked or timer expired
						this.addingMode = 2;
					} else {
						this.addingMode = 1;
					}
				}
			    this.draw();
			}
		}
		this.dragging = false;
		this.scrollmode = false;
	}
	// handles mousedown
	mousedown(event) {
		var canX = event.pageX - this.left;
		var canY = event.pageY - this.top;
		this.xmousedown = canX;
		this.ymousedown = canY;
		this.scrollYatdown = this.scrollY;
		this.dragging = true;
		this.timeDown = (new Date).getTime();
	}
	// handles mousemove
	mousemove(event) {
		var canX = event.pageX - this.left;
		var canY = event.pageY - this.top;
		
		if(this.dragging) {
			if(((this.xmousedown - canX) * (this.xmousedown - canX) +
			    (this.ymousedown - canY) * (this.ymousedown - canY)) > 9) {
			    this.scrollmode = true;
		    }
		}

		if(this.scrollmode) {
			this.scrollY = this.scrollYatdown + this.ymousedown - canY;
			// limits scrolling to usefull sizes
			if(this.scrollY > elementTypes.length * this.DN4.config.controlMenuSpace -
			                  this.height) {
			       this.scrollY = elementTypes.length * this.DN4.config.controlMenuSpace -
			                  this.height};
			if(this.scrollY < 0) {this.scrollY = 0};
		    this.draw();
		} else {
			var eh = Math.floor((canY + this.DN4.config.controlMenuSize / 2 + this.scrollY) / 
								this.DN4.config.controlMenuSpace - 1);
			if((eh < 0) || (eh >= elementTypes.length) ||
			    (canX <  (this.DN4.config.controlMenuSize - this.DN4.config.controlMenuSpace) / 2) ||
			    (canX >= (this.DN4.config.controlMenuSize + this.DN4.config.controlMenuSpace) / 2)) {
				eh = -1;
			}
	
			if(this.hoveringEl != eh) {
				this.hoveringEl = eh;
				if(eh >= 0) {
					this.DN4.draw.toolTipText = elementTypes[eh].tt;
				} else {
					this.DN4.draw.toolTipText = "";
				}
			    this.draw();
				this.DN4.draw.drawView();
			}
		}
	}
	// handles mouseout
	mouseout() {
		this.dragging = false;
		this.scrollmode = false;

		if(this.hoveringEl >= 0) {
			this.hoveringEl = -1;
			this.DN4.draw.toolTipText = "";
		    this.draw();
			this.DN4.draw.drawView();
		}
	}
}
