"use strict"	// temporary - for a better debugging

// this function encapsulates the class of drawing part of the GUI
class DN4_draw {
	constructor(tool) {
	    // all the "closure" variables
		this.DN4 = tool;			// the variable to access all the other components of the gui
	
		// html related
	    this.canvas = document.getElementById("mainCanvas");	// the canvas component
	    this.context = this.canvas.getContext("2d");			// the drawing context of the canvas
	    
	    ExtraGraph.init(this.canvas, this.context, tool);		// initializes the graphics support library
	    
	    this.width  = this.canvas.width  =
	              $(window).width() - this.DN4.config.controlMenuSize;	// height and width of the window
	    this.height = this.canvas.height = $(window).height() - this.DN4.config.mainMenuSize;
	    this.top    = this.DN4.config.mainMenuSize;
	    this.left   = 0;
	
		$("#mainCanvas").offset({ top: this.top, left: this.left})
	
		// gui related vaiables
		this.xcorner = 0;		// top-left corner
		this.ycorner = 0;
		this.xcornerdown = 0;	// top-left corner when the mouse was pressed
		this.ycornerdown = 0;
		this.xmousedown = 0;		// position where the mouse was pressed;
		this.ymousedown = 0;
		this.zoom = 1;			// current zoom
		this.zoom_down = 1;		// zoom when the mouse was pressed
		// for the grid
		this.gridSize = tool.config.initialGridSize;	// start with the initial grid size
		
		// selection related variables
		this.underSelection = new Array();		// elements under the mouse at click time
		this.currentSelection = new Array();	// elements currently selected
		this.justEdgesSelected = true;			// true if the selection includes only edges
		this.nonEdgesSelected = new Array();	// non-edges that are currently selected
		this.selboxX = 0;						// X of the selection box
		this.selboxY = 0;						// Y of the selection box
		this.selboxImg = {};					// the previous image under the rectangle
		this.prevHovering = false;				// elements over which the mouse was overing
		
		// manipulator stuff
		this.groupManType = 0;					// current type of group manipulation:
												// 0 - all, 1 - pos, 2 - rot+scal, 3 - set pivot
		this.grpMan_tx = 0;						// Group manipulator center
		this.grpMan_ty = 0;
		this.grpMan_lx = 10;					// Group manipulator size
		this.grpMan_ly = 10;
		this.grpMan_rot = 0;					// Group manipulator rotation
		this.grpMan_cx = 0;						// Group manipulator pivot point
		this.grpMan_cy = 0;		
		this.hoveringManipX = 0;				// hovering manipulator X
		this.hoveringManipY = 0;				// hovering manipulator Y
		this.hoveringManipSize = -1;			// hovering manipulator size
		this.hoveringManipNum = -1;				// hovering manipulator number
		this.hoveringManipEl  = false;			// hovering manipulator element (for edges)
		this.prevHoveringManipNum = -1;			// hovering manipulator number
		
		// Inner class for group manipulators
		class GroupManipulators extends ExtraGraph.ManipulatorHandlers {
			constructor(tool) {
				super(tool);
			}
			start(x, y, shift, manip) {	// when the group manipulator is selected
				var draw = this.DN4.draw;
				// stores the initial value of the group manipulator
				manip.grpMan_tx  = draw.grpMan_tx;						// Group manipulator center
				manip.grpMan_ty  = draw.grpMan_ty;
				manip.grpMan_lx  = draw.grpMan_lx;						// Group manipulator size
				manip.grpMan_ly  = draw.grpMan_ly;
				manip.grpMan_rot = draw.grpMan_rot;						// Group manipulator rotation
		
				switch(manip.n) {
				  case 9:			// rotation
					var draw = this.DN4.draw;
					var pivX = draw.grpMan_tx + draw.grpMan_cx * draw.grpMan_lx * Math.cos(draw.grpMan_rot) -
											    draw.grpMan_cy * draw.grpMan_ly * Math.sin(draw.grpMan_rot);
					var pivY = draw.grpMan_ty + draw.grpMan_cy * draw.grpMan_ly * Math.cos(draw.grpMan_rot) +
												draw.grpMan_cx * draw.grpMan_lx * Math.sin(draw.grpMan_rot);
				  	manip.initAng = Math.atan2(x - pivX, pivY - y);
				  	manip.initRot = draw.grpMan_rot;
				  case 3:			// x scale
				  case 5:
					manip.xSize = draw.grpMan_lx;
					manip.delta = 7;
					manip.xDirX = Math.cos(draw.grpMan_rot) * (manip.n == 3 ? -1 : 1);
					manip.xDirY = Math.sin(draw.grpMan_rot) * (manip.n == 3 ? -1 : 1);
					break;
				  case 1:			// y scale
				  case 7:
					manip.ySize = draw.grpMan_ly;
					manip.delta = 7;
					manip.yDirX = -Math.sin(draw.grpMan_rot) * (manip.n == 1 ? -1 : 1);
					manip.yDirY =  Math.cos(draw.grpMan_rot) * (manip.n == 1 ? -1 : 1);
					break;
				  case 0:			// x & y scale
				  case 2:
				  case 6:		
				  case 8:
					manip.delta = 7;
					manip.xSize = draw.grpMan_lx;
					manip.ySize = draw.grpMan_ly;
					manip.ratio = draw.grpMan_lx / draw.grpMan_ly;
					manip.xDirX =  Math.cos(draw.grpMan_rot) * ((manip.n == 0) || (manip.n == 6) ? -1 : 1);
					manip.xDirY =  Math.sin(draw.grpMan_rot) * ((manip.n == 0) || (manip.n == 6) ? -1 : 1);
					manip.yDirX = -Math.sin(draw.grpMan_rot) * ((manip.n == 0) || (manip.n == 2) ? -1 : 1);
					manip.yDirY =  Math.cos(draw.grpMan_rot) * ((manip.n == 0) || (manip.n == 2) ? -1 : 1);
					break;
				}
				if(manip.n < 10) {	// undo is done only for manipulators <10 (move, rotate or scale)
					this.DN4.cmdHistory.setCurrUndo(draw.createGrpUndo(true));
				} else if(manip.n == 11) { // undo for the move pivot
					var gmp = {cmd: "g"};
					draw.saveGroupManipulators(gmp);
					this.DN4.cmdHistory.setCurrUndo([gmp]);
				}
			}
			move(x, y, shift, manip) {	// when the group manipulator is moved
				var draw = this.DN4.draw;
				var pivX = draw.grpMan_tx + draw.grpMan_cx * draw.grpMan_lx * Math.cos(draw.grpMan_rot) -
										    draw.grpMan_cy * draw.grpMan_ly * Math.sin(draw.grpMan_rot);
				var pivY = draw.grpMan_ty + draw.grpMan_cy * draw.grpMan_ly * Math.cos(draw.grpMan_rot) +
											draw.grpMan_cx * draw.grpMan_lx * Math.sin(draw.grpMan_rot);
				
				var adjPivot = function() {
					if(draw.groupManType != 2) {	// not do it if only rotation and scale of elements is considered
						var NpivX = draw.grpMan_tx + draw.grpMan_cx * draw.grpMan_lx * Math.cos(draw.grpMan_rot) -
												     draw.grpMan_cy * draw.grpMan_ly * Math.sin(draw.grpMan_rot);
						var NpivY = draw.grpMan_ty + draw.grpMan_cy * draw.grpMan_ly * Math.cos(draw.grpMan_rot) +
													 draw.grpMan_cx * draw.grpMan_lx * Math.sin(draw.grpMan_rot);
					  	draw.grpMan_tx += pivX - NpivX;
					  	draw.grpMan_ty += pivY - NpivY;
					}
				};
				switch(manip.n) {
				  case 4:		// translation
//// SNAP SUPPORT ?
				  	x = ExtraGraph.snap(x);
				  	y = ExtraGraph.snap(y);
//// SNAP SUPPORT ? (END)
				  	draw.grpMan_tx = x;
				  	draw.grpMan_ty = y;
				  	break;
				  case 9:			// rotation
//// SNAP SUPPORT ?
					var ang = Math.atan2(x - pivX, pivY - y);
				  	var newRot = manip.initRot + ang - manip.initAng;
					draw.grpMan_rot = ExtraGraph.snapAngle1(newRot);
//// SNAP SUPPORT ? (END)
					adjPivot();
					break;
				  case 3:			// x scale
				  case 5:
				  	var len = ExtraGraph.snap((x - draw.grpMan_tx) * manip.xDirX + (y - draw.grpMan_ty) * manip.xDirY - manip.delta);
				  	if(len > 0) {
				  		draw.grpMan_lx = len;
				  	}
					adjPivot();
				  	break;
				  case 1:			// x scale
				  case 7:
				  	var len = ExtraGraph.snap((x - draw.grpMan_tx) * manip.yDirX + (y - draw.grpMan_ty) * manip.yDirY - manip.delta);
				  	if(len > 0) {
				  		draw.grpMan_ly = len;
				  	}
					adjPivot();
				  	break;
				  case 0:			// x & y scale
				  case 2:
				  case 6:		
				  case 8:
				  	var lenx = ExtraGraph.snap((x - draw.grpMan_tx) * manip.xDirX + (y - draw.grpMan_ty) * manip.xDirY - manip.delta);
				  	var leny = ExtraGraph.snap((x - draw.grpMan_tx) * manip.yDirX + (y - draw.grpMan_ty) * manip.yDirY - manip.delta);
				  	if(shift) {
					  	if(lenx > 0) {
					  		draw.grpMan_lx = lenx;
					  		draw.grpMan_ly = lenx / manip.ratio;
					  	}
				  	} else {
					  	if(lenx > 0) {
					  		draw.grpMan_lx = lenx;
					  	}
					  	if(leny > 0) {
					  		draw.grpMan_ly = leny;
					  	}
				  	}
					adjPivot();
				  	break;
				  case 11:		// setting the pivot point
//// SNAP SUPPORT ?
				  	x = ExtraGraph.snap(x);
				  	y = ExtraGraph.snap(y);
//// SNAP SUPPORT ? (END)
				  	var prpXt = x - draw.grpMan_tx;	// invert translation
				  	var prpYt = y - draw.grpMan_ty;
				  	var prpXr = prpXt * Math.cos(draw.grpMan_rot) + prpYt * Math.sin(draw.grpMan_rot);	// invert rot.
				  	var prpYr = prpYt * Math.cos(draw.grpMan_rot) - prpXt * Math.sin(draw.grpMan_rot);
				  	draw.grpMan_cx = prpXr / draw.grpMan_lx;	// invert scaling
				  	draw.grpMan_cy = prpYr / draw.grpMan_ly;
		//alert(x + " " + y + " " + prpXt + " " + prpYt + " " + prpXr + " " + prpYr + " " + draw.grpMan_cx + " " + draw.grpMan_cy);
				  	break;
				}
				// check if enabled, and do continous update
				if(this.DN4.config.continuousUpdateLevel == 2) {		
			    	this.DN4.cmdHistory.execWithHistory(this.DN4.draw.applyGrpManipulator(manip));
		
					manip.grpMan_tx  = draw.grpMan_tx;						// Group manipulator center
					manip.grpMan_ty  = draw.grpMan_ty;
					manip.grpMan_lx  = draw.grpMan_lx;						// Group manipulator size
					manip.grpMan_ly  = draw.grpMan_ly;
					manip.grpMan_rot = draw.grpMan_rot;						// Group manipulator rotation
				}
			}
			end(x, y, shift, manip) {	// when the group manipulator is end
				var draw = this.DN4.draw;
				switch(manip.n) {
				  case 0:
				  case 1:
				  case 2:
				  case 3:
				  case 4:
				  case 5:
				  case 6:
				  case 7:
				  case 8:
				  case 9:
				  	// if manipulators are updated only at the end
					if(this.DN4.config.continuousUpdateLevel < 2) {
				    	this.DN4.cmdHistory.execWithHistory(draw.applyGrpManipulator(manip));
					}
				    break;
				  case 10:		// mode change
				    draw.groupManType = (shift ? (draw.groupManType + 3) :	// 3 = 4-1
				    									  (draw.groupManType + 1) ) % 4;
					draw.toolTipText = draw.hoveringManip.handlers.getManipulatorTooltip(draw.hoveringManipEl, draw.hoveringManipNum);

					draw.drawView();
				    break;
				}
				if(manip.n < 10) {	// undo is done only for manipulators <10 (move, rotate or scale)
					switch(manip.n) {
					  case 0:
					  case 1:
					  case 2:
					  case 3:
					  case 5:
					  case 6:
					  case 7:
					  case 8:
						this.DN4.cmdHistory.putInHistory("Resize group");
						break;
					  case 4:
						this.DN4.cmdHistory.putInHistory("Move group");
					  	break;
					  case 9:
						this.DN4.cmdHistory.putInHistory("Rotate group");
					  	break;
					}
				} else if(manip.n == 11) { // for the move pivot
					var gmp = {cmd: "g"};
					draw.saveGroupManipulators(gmp);
					this.DN4.cmdHistory.setCurrRedo([gmp]);
					this.DN4.cmdHistory.putInHistory("Move pivot");
					
				}
				this.DN4.menu.update();
			}
			// enters the advanced manipulators (Group)
			enterAdvanced(manip) {
				return false;
			}
			getManipulatorTooltip(el, manip) {
				var out = "";
				switch(manip) {
				  case 4:
//					out = "Drag: Move, + SHIFT: Single Direction";
					out = "Drag: Move";
					break;
				  case 1:
				  case 7:
					out = "Drag: Vertical scaling";
					break;
				  case 3:
				  case 5:
					out = "Drag: Horizontal scaling";
					break;
				  case 0:
				  case 2:
				  case 6:
				  case 8:
					out = "Drag: Free Scaling, + SHIFT: Proportional";
					break;
				  case 9:
					out = "Drag: Rotate";
					break;
				  case 10:
					out = "Mode: ";
					switch(this.DN4.draw.groupManType) {
					  case 0:
						out = out + "Transform everything";
						break;
					  case 1:
						out = out + "Transform position, leaving elements untouched";
						break;
					  case 2:
						out = out + "Transform rotation and scale, leaving elements in-place";
						break;
					  case 3:
						out = out + "Set transformation pivot point";
						break;
					  default:
						out = out + this.DN4.draw.groupManType + " unknown transform mode";
						break;
					}
					break;
				  case 11:
					out = "Move Pivot";
					break;
				  default:
					out = "Error!";
					break;
				}
				return out;	
			}				
		}
		// creates the manipulators for the group selection
		this.groupManipulators = new GroupManipulators(this.DN4);
		
		// manipulator class
		this.Manipulator = class  {
			constructor(x, y, size, n, el, handlers) {
				this.x = x;			// manipulator position
				this.y = y;
				this.size = size;	// manipulator size
				this.n = n;			// manipulator number for the corresponding element
				this.el = el;		// element being manipulated
				this.handlers = handlers;	// manipulator event handlers
			}
		}

		this.currentManipulators = new Array();	// manipulators currently displayed
		this.underManipulation = {};			// manipulators being moved
			
	
		// advancede manipulators support -- the one that appears when a manip. is kept pressed
		this.enteringAdvancedManipulationMenu = false;	// to allow mouse up before entering the menu
		this.dontShowAdvancedManipulationMenu = false;	// to allow right-click to show the menu
		this.advancedManipulatorMenu = false;			// menu to display, or false to not show it
		this.advancedManipulatroMenuSelected = -1;		// highlight menu element
			
		// add intermediate point support
		this.addPointHovering = false;
		this.addPointHoveringX = 0;
		this.addPointHoveringY = 0;
	
		// to support the drawing of new elements
		this.newElementType = -1;	// >= 0 -- The type of the element to be added
									//   -1 -- Nothing will be displayed
									//   -2 -- A cross will be displayed
		// for new element of metatype node
		this.newNodeX = 0;
		this.newNodeY = 0;
	
		// for new element of metatype edge
		this.newEdgeX1  = 0;				// coordinates of the last segment
		this.newEdgeY1  = 0;
		this.newEdgeXc1 = 0;
		this.newEdgeYc1 = 0;
		this.newEdgeXc2 = 0;
		this.newEdgeYc2 = 0;
		this.newEdgeX2  = 0;
		this.newEdgeY2  = 0;
		this.newEdge1stTangent = false;		// true only for the first tangent of the first point
		this.newEdge2ndTangent = false;		// true for the tangents to the other points
		this.newEdgePoints = [];			// the points of the edge
		
		this.newEdgeFrom = -1;		// from and to for a new edge
		this.newEdgeTo   = -1;		// from and to for a new edge
		
		this.fromMeta = false;		// to handle meta in edge inserting
	
		// text to describe the current mode
		this.currentModeText = "";
		this.toolTipText = "";
		
		// model related variables
		this.sceneTree = new RTree();	
	}
	// creates undo for all the elements in the group
	createGrpUndo(manipUndo) {
		var out = [];
		var i, j, el, k;
		
		// creates to the selected elements
		k = 0;
		for(i = 0; i < this.currentSelection.length; i++) {	// show the manipulators for all of them
			var sel = this.currentSelection[i];	// the selected element
			var selet = elementTypes[sel.t];	// the selected element type
			out[k] = this.DN4.metaTypeElements[selet.t].makeUndoObj(sel);
			k++;
			if(selet.t == "n") {	// nodes includes also all the connected edges
				for(j = 0; j < sel.s.in.length; j++) {
					el = graph[sel.s.in[j]];
					if(!el.s.selected) {
						out[k] = this.DN4.metaTypeElements[elementTypes[el.t].t].
									 makeUndoObj(el);
						k++;
					}
				}
				for(j = 0; j < sel.s.out.length; j++) {
					el = graph[sel.s.out[j]];
					if(!el.s.selected) {
						out[k] = this.DN4.metaTypeElements[elementTypes[el.t].t].
									 makeUndoObj(el);
						k++;
					}
				}
			}
		}
		if(manipUndo) {
			var gmp = {cmd: "g"};
			this.saveGroupManipulators(gmp);
			out[out.length] = gmp;
		}		
		return out;
	}
	// applies group manipulation
	applyGrpManipulator(manip) {
		var i;
		var deMan = {};
		var draw = this.DN4.draw;
				
		// compute transformation values
		deMan.d_tx  = draw.grpMan_tx  - manip.grpMan_tx;
		deMan.d_ty  = draw.grpMan_ty  - manip.grpMan_ty;
		deMan.d_rot = draw.grpMan_rot - manip.grpMan_rot;
		deMan.d_lx  = (manip.grpMan_lx == 0) ? 1 : draw.grpMan_lx  / manip.grpMan_lx;
		deMan.d_ly  = (manip.grpMan_ly == 0) ? 1 : draw.grpMan_ly  / manip.grpMan_ly;
		deMan.cx    = manip.grpMan_tx;
		deMan.cy    = manip.grpMan_ty;
		deMan.crot  = manip.grpMan_rot;

//console.log(deMan.d_lx + " " + draw.grpMan_lx + " " + manip.grpMan_lx + " >< " + deMan.d_ly + " " + draw.grpMan_ly + " " + manip.grpMan_ly);
		
		// apply transform to the selected elements
		var out = [];
		for(i = 0; i < this.currentSelection.length; i++) {	// show the manipulators for all of them
			var sel = this.currentSelection[i];	// the selected element
			var selet = elementTypes[sel.t];	// the selected element type
			var redCmds = this.DN4.metaTypeElements[selet.t].applyManip(deMan, sel);
			out = out.concat(redCmds);
		}
		var gmp = {cmd: "g"};
		this.saveGroupManipulators(gmp);
		out[out.length] = gmp;

		return out;
	}
	
	
	// draws the screen
	draw() {
	  var context = this.context; // quick access to the context
		
	  var i;
		// The shown view port is at:
		//     [x,y,w,h] = [this.xcorner / this.zoom, this.ycorner / this.zoom,
		//                  this.width / this.zoom, this.height / this.zoom]
	  var vn = this.sceneTree.search({x: this.xcorner / this.zoom, y: (this.ycorner + this.top) / this.zoom,
	                         w: this.width   / this.zoom, h: this.height  / this.zoom});
	  
	  var et;	// element type of the current element
	  for(i = 0; i < vn.length; i++) {
		et = elementTypes[vn[i].t];		// get the element type
		this.DN4.metaTypeElements[et.t].draw(context, vn[i].s, et.d); // display it, according to its meta-type
//show bounding box, for debug
//context.strokeRect(vn[i].bbox.x, vn[i].bbox.y, vn[i].bbox.w, vn[i].bbox.h);
	  }
	}

	// draw vieport: zoom and pans the screen
	drawView() {
		var context = this.context; // quick access to the context
		var i;
		
		// clears the screen
		ExtraGraph.clear();
		// applies pan and zoom
		context.save();
	    context.translate(-this.xcorner, -this.ycorner - this.top);
		context.scale(this.zoom,this.zoom);
		// shows the grid (if selected)
		if(this.DN4.menu.items.mi_grid.selected) {
			var gridExtentXmin = this.xcorner / this.zoom;
			var gridExtentYmin = (this.ycorner + this.top) / this.zoom;
			var gridExtentXmax = gridExtentXmin + this.width   / this.zoom;
			var gridExtentYmax = gridExtentYmin + this.height  / this.zoom;
			
			var gridExtentXminQ = Math.ceil(gridExtentXmin / this.gridSize) * this.gridSize;
			var gridExtentYminQ = Math.ceil(gridExtentYmin / this.gridSize) * this.gridSize;
			var gridExtentXmaxQ = Math.floor(gridExtentXmax / this.gridSize) * this.gridSize;
			var gridExtentYmaxQ = Math.floor(gridExtentYmax / this.gridSize) * this.gridSize;

			var step = this.gridSize;
			var maxLines = (gridExtentXmaxQ - gridExtentXminQ > gridExtentYmaxQ - gridExtentYminQ ?
							gridExtentXmaxQ - gridExtentXminQ : gridExtentYmaxQ - gridExtentYminQ) / this.gridSize;
			while(maxLines > 500) {
				step *= 2;
				maxLines /= 2;
			}
			context.save();
			context.beginPath();
			for(i = gridExtentXminQ; i <= gridExtentXmaxQ; i += step) {
				context.moveTo(i, gridExtentYmin);
				context.lineTo(i, gridExtentYmax);
			}
			for(i = gridExtentYminQ; i <= gridExtentYmaxQ; i += step) {
				context.moveTo(gridExtentXmin, i);
				context.lineTo(gridExtentXmax, i);
			}
			context.strokeStyle = "#555555";	//  outline
			context.lineWidth = 0.5 / this.zoom;
			context.stroke();
			context.restore();
		}
		// draws the visible area
		this.draw();

		// display where the new node will be inserted
		if(this.newElementType >= 0) {	// there is a preview of an element to add
			var et = elementTypes[this.newElementType];
			if(et.t == 'n') {	// for nodes
				this.DN4.metaTypeElements[et.t].draw(context,
				    {
				    	tx: this.newNodeX,
				    	ty: this.newNodeY,
				    	sx: 1.0, sy:1, rot: 0, in: [], out: []
				    }, et.d);
			} else if(et.t == 'e') {	// for edges
				if(this.newEdge1stTangent) {	// the first tangent is handled differently
					context.save();
					context.moveTo(this.newEdgeX1,  this.newEdgeY1);
					context.lineTo(this.newEdgeXc1, this.newEdgeYc1);
					context.strokeStyle = "#999999";	//  outline
					context.lineWidth = 1;
					context.stroke();
					context.restore();
				} else {
					i = this.newEdgePoints.length-1;
					this.newEdgePoints[i] = {
						x1:  this.newEdgeX1,  y1:  this.newEdgeY1,
						xc1: this.newEdgeXc1, yc1: this.newEdgeYc1,
						t1: 7,
						xc2: this.newEdgeXc2, yc2: this.newEdgeYc2,
						x2:  this.newEdgeX2,  y2:  this.newEdgeY2,
						t2: 7
					};
					if(this.newEdge2ndTangent) {	// adds a guide for other tangents
						context.save();
						context.moveTo(2 * this.newEdgeX2 - this.newEdgeXc2,
									   2 * this.newEdgeY2 - this.newEdgeYc2);
						context.lineTo(this.newEdgeXc2, this.newEdgeYc2);
						context.strokeStyle = "#999999";	//  outline
						context.lineWidth = 1;
						context.stroke();
						context.restore();
					}
//alert(this.newEdgePoints.length);
					this.DN4.metaTypeElements[et.t].draw(context,
					    {
					    	from:   this.newEdgeFrom,
					    	to:     -1,
					    	points: this.newEdgePoints
					    }, et.d);
				}
			}
		} else if (this.newElementType == -2) {
			context.save();
			context.lineCap = "round";
			context.beginPath();
			context.moveTo(this.newNodeX - 7, this.newNodeY - 7);
			context.lineTo(this.newNodeX + 7, this.newNodeY + 7);
			context.moveTo(this.newNodeX - 7, this.newNodeY + 7);
			context.lineTo(this.newNodeX + 7, this.newNodeY - 7);
			context.strokeStyle = "#ffffff";	//  outline
			context.lineWidth = 7;
			context.stroke();
			context.strokeStyle = "#ff0000";	//  cross
			context.lineWidth = 3;
			context.stroke();
			context.restore();
		}

		// display manipulators for selected objects
		// temporary: it must be done only when something changes - now it is done everytime
		this.currentManipulators = new Array();	// manipulators currently displayed

		if(this.currentSelection.length == 1) {	// if there is a single object selected, show its manipulators
			var sel = this.currentSelection[0];	// the selected element
			var selet = elementTypes[sel.t];	// the selected element type
			this.DN4.metaTypeElements[selet.t].drawManipulators(context, sel);	// display manipulators
		} else if(this.currentSelection.length > 1) { // if there are more than one object selected
			if(this.justEdgesSelected) {	// if only edges are selected
				for(i = 0; i < this.currentSelection.length; i++) {	// show the manipulators for all of them
					var sel = this.currentSelection[i];	// the selected element
					var selet = elementTypes[sel.t];	// the selected element type
					this.DN4.metaTypeElements[selet.t].drawManipulators(context, sel);
				}
			} else {	//  otherwise, show the group manipulator
				// draws the manipulators
				context.strokeStyle = "#0077ff";	//  sets the color
				context.fillStyle   = "#ffffff";	//  sets the color
				
				// draw the pivot cross
				var pivX = this.grpMan_tx + this.grpMan_cx * this.grpMan_lx * Math.cos(this.grpMan_rot) -
										    this.grpMan_cy * this.grpMan_ly * Math.sin(this.grpMan_rot);
				var pivY = this.grpMan_ty + this.grpMan_cy * this.grpMan_ly * Math.cos(this.grpMan_rot) +
											this.grpMan_cx * this.grpMan_lx * Math.sin(this.grpMan_rot);
				context.save();
				context.translate(pivX, pivY);
				context.rotate(this.grpMan_rot);
				context.beginPath();
				context.moveTo(-11, 0);
				context.lineTo( 11, 0);
				context.moveTo(0, -11);
				context.lineTo(0,  11);
				context.stroke();
				context.restore();
				
				if(this.groupManType <= 2) {	// only for all, pos and r/s type
					ExtraGraph.drawManipulatorRect(this.grpMan_tx, this.grpMan_ty, this.grpMan_rot,
												this.grpMan_lx, this.grpMan_ly,
												{}, this.groupManipulators, this);
				} else { // otherwise draws only the rectangle
					// draw the pivot manipulator
			 		if(this.hoveringManipNum == 11) {	// for hovering
			 			context.save();
						context.fillStyle   = "#ff0000";
			 		}
			 		context.fillRect(  pivX - 3,  pivY - 3,  7, 7);
					if(this.hoveringManipNum == 11) {	// for hovering
						context.restore();
					}
					context.strokeRect(pivX - 3,  pivY - 3,  7, 7);
					this.addManipulator(pivX, pivY,
										4, 11, {}, this.groupManipulators);
					// draw the element bounding box
					context.translate(this.grpMan_tx, this.grpMan_ty);
					context.rotate(this.grpMan_rot);
				  	context.strokeRect(-this.grpMan_lx, -this.grpMan_ly,
				  						this.grpMan_lx * 2, this.grpMan_ly * 2);
				}
		 		if(this.hoveringManipNum == 10) {	// for hovering
		 			context.save();
					context.fillStyle   = "#ff0000";
		 		}
		 		context.fillRect(  -this.grpMan_lx -10,  this.grpMan_ly +17,  7, 7);
				if(this.hoveringManipNum == 10) {	// for hovering
					context.restore();
				}
				context.strokeRect(-this.grpMan_lx -10,  this.grpMan_ly +17,  7, 7);
		 		context.fillRect(-this.grpMan_lx,  this.grpMan_ly + 12,  24, 18);
				ExtraGraph.centerText(12-this.grpMan_lx, this.grpMan_ly
													    +this.DN4.config.textBaseSize * 2.5, 
						this.groupManType == 0 ? "all" :
						this.groupManType == 1 ? "pos" : 
						this.groupManType == 2 ? "r/s" : "(-)",
						"#0077ff", this.DN4.config.textBaseSize);
				this.addManipulator(this.grpMan_tx - (this.grpMan_ly + 20) * Math.sin(this.grpMan_rot)
												   - (this.grpMan_lx + 8)  * Math.cos(this.grpMan_rot), 
									this.grpMan_ty + (this.grpMan_ly + 20) * Math.cos(this.grpMan_rot)
												   - (this.grpMan_lx + 8)  * Math.sin(this.grpMan_rot),
									4, 10, {}, this.groupManipulators);
			}
		}

		// display context menus for advanced manipulators
		if(this.advancedManipulatorMenu) {
			context.fillStyle   = "#ffdddd";
			context.fillRect(this.advancedManipulatorMenu.x - 3, 
							 this.advancedManipulatorMenu.y - this.DN4.config.textBaseSize*1.5 - 3,
							 this.DN4.config.textBaseSize * this.advancedManipulatorMenu.w,
							 this.DN4.config.textBaseSize * 1.5 * 
							      this.advancedManipulatorMenu.items.length+ 7);

			for(i = 0; i < this.advancedManipulatorMenu.items.length; i++) {
				ExtraGraph.leftText(this.advancedManipulatorMenu.x,
									this.advancedManipulatorMenu.y +
									     this.DN4.config.textBaseSize * 1.5 * i,
									this.advancedManipulatorMenu.items[i].label,
									this.advancedManipulatroMenuSelected == i ? "#ff0000" : "#0077ff",
									this.DN4.config.textBaseSize);
			}
//alert(this.advancedManipulatorMenu.x + " " + this.advancedManipulatorMenu.y)
//			context.fillRect(this.advancedManipulatorMenu.x-10,
//							 this.advancedManipulatorMenu.y-10,
//							 21, 21);
		}
		
		// display position of inserted point in add intermediate point mode
		if(this.addPointHovering) {
			context.strokeStyle = "#000000";	//  sets the color
			context.fillStyle   = "#ffffff";	//  sets the color
			context.beginPath();
			context.arc(this.addPointHoveringX, this.addPointHoveringY, 7, 0, 2*Math.PI);
			context.fill();
			context.stroke();
		}

		// resotres the standard systmem of coordinates
		context.restore();
		
		// display some HUD
		ExtraGraph.centerText(this.width/2, this.height-2,
					"ALT + click: pan view,    CTRL + click: zoom view", "black",
					Math.round(this.DN4.config.textBaseSize * 1.33));
		
		if(this.toolTipText != "") {
			ExtraGraph.centerText(this.width/2, this.DN4.config.textBaseSize * 2.33,
					this.toolTipText, "black",
					Math.round(this.DN4.config.textBaseSize * 1.33));
		} else if(this.currentModeText != "") {
			ExtraGraph.centerText(this.width/2, this.DN4.config.textBaseSize * 2.33,
					this.currentModeText, "black",
					Math.round(this.DN4.config.textBaseSize * 1.33));
		}
	}
	
	// Initialization
	init() {
	  var i;
	// fill rTree
		  
	  // inserts the elements in the rTree
	  for(i = 0; i < graph.length; i++) {
	  	var g = graph[i];
	  	if(g.t >= 0) { // inserts only proper elements
		  	this.sceneTree.insert(g.bbox, g);
	  	}
	  }
	  this.drawView();
	};

	// resizes the window
	resize() {
	    this.width = this.canvas.width = $(window).width() - this.DN4.config.controlMenuSize;
	    this.height = this.canvas.height = $(window).height() - this.DN4.config.mainMenuSize;
		$("#mainCanvas").offset({ top: this.top, left: this.left})
	    
	    this.drawView();
	}
	
	// returns the list of elements under the current mouse point
	atPoint(canX, canY) {
		var out = new Array();

		var underMouse = this.sceneTree.search(		// search elements under the mouse
								{x: canX - 2 / this.zoom,
								 y: canY - 2 / this.zoom,
	                             w: 5 / this.zoom, h: 5 / this.zoom});
	                             
	    var i, cnt = 0;
	    var firstNode = -1;		// if it founds a node, memorize its position
	    for(i = 0; i < underMouse.length; i++) {
			var et = elementTypes[underMouse[i].t];		// get the element type
			// check if the mouse is over this element
			if(this.DN4.metaTypeElements[et.t].isOver(underMouse[i].s, et.d, canX, canY)) {
				out[cnt] = underMouse[i];
				// get the first non-edge
				if((elementTypes[underMouse[i].t].t != 'e') && (firstNode == -1)) {
					firstNode = cnt;
				}
				cnt++;
	    	}
	    }
	    if(firstNode > 0) {	// if non-edges are available, make sure that the first element is not an edge
	    	var tmp = out[0];
	    	out[0] = out[firstNode];
	    	out[firstNode] = tmp;
	    }
	    return out;
	}
	
	// returns the size of the current selection
	currentSelectionSize() {
		return this.currentSelection.length;
	}
	// clears the current selection
	clearCurrentSelection() {
		var i;
		for(i = 0; i < this.currentSelection.length; i++) {
			this.currentSelection[i].s.selected = false;
		}
		this.currentSelection = new Array();
		this.nonEdgesSelected = new Array();
		this.justEdgesSelected = true;
	}
	// creates an array from the current selection
	currentSelectionToArray() {
		var i;
		var out = [];
		
		for(i = 0; i < this.currentSelection.length; i++) {
			out[i] = this.currentSelection[i];
		}
		
		return out;
	}
	
	// remove the given element from the selection
	removeFromCurrentSelection(el, dontPropagate) {
	  if(el.s.selected) {	// only if selected
		var i, j, na;

		if(elementTypes[el.t].t != "e") {
///// 2014 - 07 - 19: i do not understand why i put this if!!!!!!!!!			
		}

		this.justEdgesSelected = true;
		j = 0;
		na = new Array();
		for(i = 0; i < this.currentSelection.length; i++) {
			if(this.currentSelection[i] === el) {	// this is the target el, so unselect it
				this.currentSelection[i].s.selected = false;
			} else {	// copy the normal elements
				// check if the remaining elements are only edges
				if(elementTypes[this.currentSelection[i].t].t != "e") {
					this.justEdgesSelected = false;
				}
				// copy the element				
				na[j] = this.currentSelection[i];
				j++;
			}
		}
		this.currentSelection = na;
		// non edges must be also removed from the nonEdgesSelected array
		if(elementTypes[el.t].t != "e") {
			j = 0;
			na = new Array();
			for(i = 0; i < this.nonEdgesSelected.length; i++) {
				if(!(this.nonEdgesSelected[i] === el)) {	// this is the target el, so remove it
					na[j] = this.nonEdgesSelected[i];
					j++;
				}
			}
			this.nonEdgesSelected = na;
		}
		
		// now check if some dependent node must be unselected
		switch(elementTypes[el.t].t) {
		  case 'n': 	// nodes removes also in and out arcs
			for(i = 0; i < el.s.in.length; i++) {
				this.removeFromCurrentSelection(graph[el.s.in[i]], true);	// note: do not propagate
			}											// this avoids to deselect the entire sub-net
			for(i = 0; i < el.s.out.length; i++) {
				this.removeFromCurrentSelection(graph[el.s.out[i]], true);
			}
			break;
		  case 'e':		// edges removes also "from" and "to"
		  	if(!dontPropagate) {	// only if the edge removal must propagate to neighbor nodes
				this.removeFromCurrentSelection(graph[el.s.from]);
				this.removeFromCurrentSelection(graph[el.s.to]);
		  	}
			break;
		}
	  }
	}
	
	// adds "from" and "to" of an edge to the current selection
	addEdgeExtremitiesToCurrentSelection(el) {
		this.addToCurrentSelection(graph[el.s.from]);
		this.addToCurrentSelection(graph[el.s.to]);
	}

	// add the given element to the selection (if unselected)
	// also adds the connected elements (for edges)
	addToCurrentSelection(el) {
		var i, l, edgEl;
		if(! el.s.selected) {	// consider only unselected elements
			// adds to the selection array, and show it as selected
			l = this.currentSelection.length;
			this.currentSelection[l] = el;
			el.s.selected = true;
			
			// if selecting is not an edge, keep it separately
			if(elementTypes[el.t].t != "e") {
				if(this.justEdgesSelected) {	// if only edges were selected before
					this.justEdgesSelected = false;
					// add their "from" and "to" to the selection
					for(i = 0; i < l; i++) {
						this.addEdgeExtremitiesToCurrentSelection(this.currentSelection[i]);
					}
				}
				// add the node
				this.nonEdgesSelected[this.nonEdgesSelected.length] = el;
				// if it is a node, check if there are edge that connects it to an element
				// already selected. If so, include also the corresponding edges
				// 1) check the input arcs
				for(i = 0; i < el.s.in.length; i++) {
					edgEl = graph[el.s.in[i]];
					if(! edgEl.s.selected) { // only if not already selected
						if(graph[edgEl.s.from].s.selected) {	// if the other extremity is selected, adds it
							this.currentSelection[this.currentSelection.length] = edgEl;
							edgEl.s.selected = true;
						}
					}
				}
				// 2) check the output arcs
				for(i = 0; i < el.s.out.length; i++) {
					edgEl = graph[el.s.out[i]];
					if(! edgEl.s.selected) { // only if not already selected
						if(graph[edgEl.s.to].s.selected) {	// if the other extremity is selected, adds it
							this.currentSelection[this.currentSelection.length] = edgEl;
							edgEl.s.selected = true;
						}
					}
				}
			} else {
				if(! this.justEdgesSelected) { // if selecting an edge, when there are also other type of elements
					// adds also the "to" and "from" nodes
					this.addEdgeExtremitiesToCurrentSelection(el);
				}
			}
		}
	}
	
	// compute the initial position / size of the group manipulator
	initGroupManipulators() {
		// works only if the current selection will need group manipulators
		if((this.currentSelection.length > 1) && (! this.justEdgesSelected)) {
			var minX, maxX, minY, maxY, i;
			// find bounding box of selected elements
			var nonEdges = this.nonEdgesSelected;
			minX = maxX = nonEdges[0].s.tx;
			minY = maxY = nonEdges[0].s.ty;
			for(i = 1; i < nonEdges.length; i++) {
				if(nonEdges[i].s.tx < minX) {minX = nonEdges[i].s.tx; }
				if(nonEdges[i].s.ty < minY) {minY = nonEdges[i].s.ty; }
				if(nonEdges[i].s.tx > maxX) {maxX = nonEdges[i].s.tx; }
				if(nonEdges[i].s.ty > maxY) {maxY = nonEdges[i].s.ty; }
	
			}
			// inits the group manipulator variables
			this.grpMan_tx = (minX + maxX) / 2;
			this.grpMan_ty = (minY + maxY) / 2;
			this.grpMan_lx = (maxX - minX) / 2;
			this.grpMan_ly = (maxY - minY) / 2;
			this.grpMan_rot = 0;
			this.grpMan_cx = 0;
			this.grpMan_cy = 0;
		}
	}
	// save the group manipulator parameters
	saveGroupManipulators(el) {
		el.tx  = this.grpMan_tx;
		el.ty  = this.grpMan_ty;
		el.lx  = this.grpMan_lx;
		el.ly  = this.grpMan_ly;
		el.rot = this.grpMan_rot;
		el.cx  = this.grpMan_cx;
		el.cy  = this.grpMan_cy;
	}
	// restore the group manipulator parameters
	restoreGroupManipulators(el) {
		this.grpMan_tx  = el.tx;
		this.grpMan_ty  = el.ty;
		this.grpMan_lx  = el.lx;
		this.grpMan_ly  = el.ly;
		this.grpMan_rot = el.rot;
		this.grpMan_cx  = el.cx;
		this.grpMan_cy  = el.cy;
	}
	// tells if group manipulators are shown
	areGroupManipulatorsShown() {
		return ((this.currentSelection.length > 1) && (!this.justEdgesSelected));
	}

	// returns the list of elements under the current mouse point
	selBox(canX, canY, canW, canH) {
		var out = new Array();

		var underMouse = this.sceneTree.search(		// search elements under the mouse
								{x: canX, y: canY, w: canW, h: canH});
	                             
	    var i, cnt = 0;
	    for(i = 0; i < underMouse.length; i++) {
			var et = elementTypes[underMouse[i].t];		// get the element type
			// check if the mouse is over this element
			if(this.DN4.metaTypeElements[et.t].selBox(underMouse[i].s, et.d, canX, canY, canW, canH)) {
				out[cnt] = underMouse[i];
				cnt++;
	    	}
	    }
	    return out;
	}
	
	// adds a manipulator to the current array of manipulators
	addManipulator(x, y, size, n, el, handlers) {
		this.currentManipulators[this.currentManipulators.length] =
				new this.Manipulator(x, y, size, n, el, handlers);
	}
	
	// check if there is a manipulator at the given position. If so, starts the msnipulation
	// and returns true. Otherwise it returns false
	manipulatorAtPoint(canX, canY, shift) {
		var i;
		for(i = 0; i < this.currentManipulators.length; i++) {
			var cm = this.currentManipulators[i];	// fast access to the considered manipulator
			if((canX >= cm.x - cm.size) && (canY >= cm.y - cm.size) &&
			   (canX <= cm.x + cm.size) && (canY <= cm.y + cm.size)) {
			   	cm.handlers.start(canX, canY, shift, cm);
			   	this.underManipulation = cm;
			   	return true;
			}
		}
		return false;
	}
	// check if the mouse is hovering over a manipulator
	hoveringManipulatorAtPoint(canX, canY) {
		var i;
		for(i = 0; i < this.currentManipulators.length; i++) {
			var cm = this.currentManipulators[i];	// fast access to the considered manipulator
			if((canX >= cm.x - cm.size) && (canY >= cm.y - cm.size) &&
			   (canX <= cm.x + cm.size) && (canY <= cm.y + cm.size)) {
				this.hoveringManip = cm;
			   	this.hoveringManipSize = cm.size;
			   	this.hoveringManipX = cm.x;
			   	this.hoveringManipY = cm.y;
				this.hoveringManipNum = cm.n;
				this.hoveringManipEl = cm.el;
			   	return true;
			}
		}
		this.hoveringManipSize = -1;
		this.hoveringManipNum = -1;
		return false;
	}
	// Align
	alignSelected(dir, axis) {
		var undoOut = this.DN4.draw.createGrpUndo(true);
		var boxMode = this.DN4.menu.items.mi_al_box.selected;
		
		var i;
//console.log("Align " + axis + " " + dir);
		var min, max, ave, val;
		var n = "t" + axis;
		min = max = this.nonEdgesSelected[0].s[n];
		for(i = 1; i < this.nonEdgesSelected.length; i++) {
			var sel = this.nonEdgesSelected[i];	// the selected element
			min = (sel.s[n] < min ? sel.s[n] : min);
			max = (sel.s[n] > max ? sel.s[n] : max);
		}
		ave = (min + max) / 2;
		val = (dir == -1 ? min : (dir == 1 ? max : ave));
//console.log(min, ave, max, val);

		var deMan = {d_tx: 0, d_ty: 0, d_rot: 0, d_lx: 1, d_ly: 1, cx: 0, cy: 0, crot: 0};
		var dn = "d_" + n;		
		// compute transformation values
		var out = [];
		// First handles non-edge selected
		for(i = 0; i < this.nonEdgesSelected.length; i++) {
			var sel = this.nonEdgesSelected[i];	// the selected element
			var selet = elementTypes[sel.t];	// the selected element type
			deMan[dn] = val - sel.s[n];
			var redCmds = this.DN4.metaTypeElements[selet.t].applyManip(deMan, sel);
			out = out.concat(redCmds);
		}
		// then handles selected edges, to match the new stating and ending points
		var k = out.length;
		for(i = 0; i < this.currentSelection.length; i++) {
			var sel = this.currentSelection[i];	// the selected element
			var selet = elementTypes[sel.t];	// the selected element type
			if(selet.t == "e") {	// Transform only edge elements
				var points = sel.s.points;
				var last = points.length-1;
				var poses = {x1 : points[0].x1, y1 : points[0].y1, x2 : points[last].x2, y2 : points[last].y2}
				poses[axis+"1"] = val;
				poses[axis+"2"] = val;
				out[k] = {cmd: "p", el: sel, path: ["points"], more: []};
				ExtraGraph.expandEdge(points, last, out[k].more, poses.x1, poses.y1, poses.x2, poses.y2);
				k++;;		
			}
		}
		var gmp = {cmd: "g"};
		out[out.length] = gmp;

		this.DN4.cmdHistory.addAndExec("Align", undoOut, out);
		this.initGroupManipulators();
		this.saveGroupManipulators(gmp);
		this.drawView();
	}
	
	// Distribute or space
	distribSelected(axis) {
		var undoOut = this.DN4.draw.createGrpUndo(true);
		var boxMode = this.DN4.menu.items.mi_al_box.selected;
		
		var i;
//console.log("Distribute " + axis);
		var min, max, idArr;
		var n = "t" + axis;
		var e0 = this.nonEdgesSelected[0];
		min = max = e0.s[n];
		idArr = [{id: e0.id, val: e0.s[n]}];
		for(i = 1; i < this.nonEdgesSelected.length; i++) {
			var sel = this.nonEdgesSelected[i];	// the selected element
			min = (sel.s[n] < min ? sel.s[n] : min);
			max = (sel.s[n] > max ? sel.s[n] : max);
			idArr[i] = {id: sel.id, val: sel.s[n]};
		}
		idArr.sort(function(a, b){return a.val-b.val});
		var Delta = (max - min) / (this.nonEdgesSelected.length - 1);
		var deMan = {d_tx: 0, d_ty: 0, d_rot: 0, d_lx: 1, d_ly: 1, cx: 0, cy: 0, crot: 0};
		var dn = "d_" + n;		
		// compute transformation values
		var out = [];
		var trad = [];	// array to remember the new positions
		// First handles non-edge selected
		for(i = 0; i < this.nonEdgesSelected.length; i++) {
			var sel = graph[idArr[i].id];	// the selected element
			var selet = elementTypes[sel.t];	// the selected element type
			var val = min + Delta*i;
			deMan[dn] = val - sel.s[n];
			trad[idArr[i].id] = val;		// remembers the new value
			var redCmds = this.DN4.metaTypeElements[selet.t].applyManip(deMan, sel);
			out = out.concat(redCmds);
		}
		// then handles selected edges, to match the new stating and ending points
		var k = out.length;
		for(i = 0; i < this.currentSelection.length; i++) {
			var sel = this.currentSelection[i];	// the selected element
			var selet = elementTypes[sel.t];	// the selected element type
			if(selet.t == "e") {	// Transform only edge elements
				var points = sel.s.points;
				var last = points.length-1;
				var poses = {x1 : points[0].x1, y1 : points[0].y1, x2 : points[last].x2, y2 : points[last].y2}
				poses[axis+"1"] = trad[sel.s.from];
				poses[axis+"2"] = trad[sel.s.to];
				out[k] = {cmd: "p", el: sel, path: ["points"], more: []};
				ExtraGraph.expandEdge(points, last, out[k].more, poses.x1, poses.y1, poses.x2, poses.y2);
				k++;;		
			}
		}
		var gmp = {cmd: "g"};
		out[out.length] = gmp;

		this.DN4.cmdHistory.addAndExec("Align", undoOut, out);
		this.initGroupManipulators();
		this.saveGroupManipulators(gmp);
		this.drawView();
	}

	// Apply an alignment function to the current selection
	// Main function codes:
	// Function   0 -> Align elements
	// Function   1 -> Distribute or space elements
	// Function  10 -> Snap selected elements to grid
	applyAlignFunction(func, sub, axis) {
// To play safe!
		alert("Function not working yet");
		return;
		
		
		var undoOut = this.DN4.draw.createGrpUndo(true);		// creates the undo (as for group manipulators)
		var boxMode = this.DN4.menu.items.mi_al_box.selected;
		var otherAxis = (axis == "x") ? "y" : "x";
		
		var i;
		
		var min, max, ave, idArr;
		var n  = "t" + axis;
		var no = "t" + otherAxis;
		if(func < 10) { // functions < 10 works on a single axis, and requires minimum and maximum
			var e0 = this.nonEdgesSelected[0];
			min = e0.s[n];
			max = e0.s[n];
			idArr = [{id: e0.id, val: e0.s[n], vo: e0.s[no]}];
			for(i = 1; i < this.nonEdgesSelected.length; i++) {
				var sel = this.nonEdgesSelected[i];	// the selected element
				var v = sel.s[n];
				var vo = sel.s[no];
				idArr[i] = {id: sel.id, val: v, vo: vo};
				var valMin = v;
				var valMax = v;
				min = (valMin < min ? valMin : min);
				max = (valMax > max ? valMax : max);
			}
		}
		var ave = (max + min) / 2;
		
		var newPos = [];	// Array of objects with new positions
		var commandName;

		console.log(axis + " " + otherAxis);
		switch(func) {
		  case 0:
		  	commandName = "Align";
			for(i = 0; i < idArr.length; i++) {
				newPos[i] = {};
				newPos[i][axis] = (sub == -1 ? min : (sub == 1 ? max : ave));
				newPos[i][otherAxis] = idArr[i].vo; 
			}
		  	break;
		  case 1:
		  	this.distribSelected(axis);
		  	break;
		  case 10:
		  	console.log("Snap to Grid");
		  	break;
		  default:
		  	console.log("Function " + func + " not implemented yet");
		  	break;
		}
		
		for(i = 0; i < idArr.length; i++) {
			console.log(i + " " + idArr[i].id + " " + newPos[i].x + " " + newPos[i].y);
// Mi ero fermato qui, perche' adesso si sarebbe dovuto generare un comando
// per poi poter supportare l'UNDO
		}
	}

	
	// Quick transform
	quickTransform(tx, ty, r, sx, sy) {
// To play safe!
		alert("Function not working yet");
		return;

		var undoOut = this.DN4.draw.createGrpUndo(true);

		var i;
		var pivX = this.grpMan_tx + this.grpMan_cx * this.grpMan_lx * Math.cos(this.grpMan_rot) -
								    this.grpMan_cy * this.grpMan_ly * Math.sin(this.grpMan_rot);
		var pivY = this.grpMan_ty + this.grpMan_cy * this.grpMan_ly * Math.cos(this.grpMan_rot) +
									this.grpMan_cx * this.grpMan_lx * Math.sin(this.grpMan_rot);
//console.log("Quick Transform ", tx, ty, r, sx, sy, pivX, pivY);
		var deMan = {d_tx: tx, d_ty: ty, d_rot: r, d_lx: sx, d_ly: sy, cx: pivX, cy: pivY, crot: 0};
		var out = [];
		for(i = 0; i < this.currentSelection.length; i++) {	// show the manipulators for all of them
			var sel = this.currentSelection[i];	// the selected element
			var selet = elementTypes[sel.t];	// the selected element type
			var redCmds = this.DN4.metaTypeElements[selet.t].applyManip(deMan, sel);
			out = out.concat(redCmds);
		}
		var gmp = {cmd: "g"};
		out[out.length] = gmp;

		this.DN4.cmdHistory.addAndExec("QuickTransf", undoOut, out);
		this.initGroupManipulators();
		this.saveGroupManipulators(gmp);
		this.drawView();
	}
	
	centerView() {
		var stree = this.sceneTree.get_tree();
		
		var bb_cx = stree.x + stree.w / 2;
		var bb_cy = stree.y + stree.h / 2;
		
		this.zoom = 1;
		this.xcorner = bb_cx - this.width / 2;
		this.ycorner = bb_cx - this.height / 2;

		
		this.drawView();
	}
}
