"use strict"	// temporary - for a better debugging

// this function encapsulates the class of drawing part of the GUI
class DN4_ExtraGraph {
	constructor() {
	    this.canvas = {};		// property to contain the canvas
	    this.context = {};		// property to contain the context
	    this.tool = {};			// property to contain the main tool object

		// for drawing SVG (taken from internet)
		this.SVGmarkerRegEx = /[MmLlSsQqLlHhVvCcSsQqTtAaZz]/g;

		// to use timers
		this.currentTimer = "none";

		// generic class for manipulator events
		// manipulator handler class
		this.ManipulatorHandlers = class {
			constructor(tool) {
				this.DN4 = tool;
			}
			// start manipulation prototype
			start(x, y, shift, manip) {
			}
			// move manipulaor prototype
			move(x, y, shift, manip) {
			}
			// end manipulation prototype
			end(x, y, shift, manip) {
			}
			// enters advanced manipulation state
			enterAdvanced(manip) {
				return false;	// returns the state for this advanced manupulators, of false if
								// no advanced manipulators need to be shown
			}
			// returns the Tool Tip for the selected manipulator
			getManipulatorTooltip(el, manip) {
				return "undefined tooltip: " + el.n + " " + manip;
			}
		}
	}
	    
    // init the ExtraGraph library
    init(_canvas, _context, _tool) {
    	this.canvas = _canvas;
    	this.context = _context;
    	this.tool = _tool;
    }

	// prints a centered text
	centerText(x, y, text, text_col, size) {
		var context = this.context; // quick access to the context
		
	    context.font = size+"pt Calibri";
	    context.fillStyle = text_col; // text color
	     
	    var metrics = context.measureText(text);
	 
	    context.fillText(text,
	        x - metrics.width / 2,
	        y - size / 2);
	}

	// prints a left aligned text
	leftText(x, y, text, text_col, size) {
		var context = this.context; // quick access to the context
		
	    context.font = size+"pt Calibri";
	    context.fillStyle = text_col; // text color
	     
	    context.fillText(text,
	        x,
	        y - size / 2);
	}

	svgPathToCommands(str) {
	    var results = []; 
	    var match; while ((match = this.SVGmarkerRegEx.exec(str)) !== null) { results.push(match); };
	    return results
	        .map(function(match) {
	            return { marker: str[match.index], 
	                     index: match.index };
	        })
	        .reduceRight(function(all, cur) {
	            var chunk = str.substring(cur.index, all.length ? all[all.length - 1].index : str.length);
	            return all.concat([
	               { marker: cur.marker, 
	                 index: cur.index, 
	                 chunk: (chunk.length > 0) ? chunk.substr(1, chunk.length - 1) : chunk }
	            ]);
	        }, [])
	        .reverse()
	        .map(function(command) {
	            var values = command.chunk.match(/-?[0-9]*\.?\d+/g);
	            return { marker: command.marker, values: values ? values.map(parseFloat) : []};
	        });
	}
	
	drawSvgPathFromCommandList(ctx, commandList) {
	    ctx.beginPath();
	    var lastPos = [ 0, 0 ]; var pointOne, pointTwo;
	    commandList.forEach(function(command) {
		    if ((command.marker === 'z') || (command.marker === 'Z')) {
		        lastPos = [ 0, 0 ];
		        ctx.closePath();
		    } else if (command.marker === 'm') {
		        lastPos = [ lastPos[0] + command.values[0], lastPos[1] + command.values[1] ];
		        ctx.moveTo(lastPos[0], lastPos[1]);
		    } else if (command.marker === 'l') {
		        lastPos = [ lastPos[0] + command.values[0], lastPos[1] + command.values[1] ];
		        ctx.lineTo(lastPos[0], lastPos[1]);
		    } else if (command.marker === 'h') {
		        lastPos = [ lastPos[0] + command.values[0], lastPos[1] ];
		        ctx.lineTo(lastPos[0], lastPos[1]);
		    } else if (command.marker === 'v') {
		        lastPos = [ lastPos[0], lastPos[1] + command.values[0] ];
		        ctx.lineTo(lastPos[0], lastPos[1]);
		    } else if (command.marker === 'c') {
		        pointOne = [ lastPos[0] + command.values[0],
		                     lastPos[1] + command.values[1] ];
		        pointTwo = [ lastPos[0] + command.values[2],
		                     lastPos[1] + command.values[3] ];
		        lastPos  = [ lastPos[0] + command.values[4],
		                     lastPos[1] + command.values[5] ];
		        ctx.bezierCurveTo(
		                pointOne[0], pointOne[1],
		                pointTwo[0], pointTwo[1],
		                lastPos[0], lastPos[1]);
		    } else if (command.marker === 'M') {
		        lastPos = [ command.values[0], command.values[1] ];
		        ctx.moveTo(lastPos[0], lastPos[1]);
		    } else if (command.marker === 'L') {
		        lastPos = [ command.values[0], command.values[1] ];
		        ctx.lineTo(lastPos[0], lastPos[1]);
		    } else if (command.marker === 'H') {
		        lastPos = [ command.values[0], lastPos[1] ];
		        ctx.lineTo(lastPos[0], lastPos[1]);
		    } else if (command.marker === 'V') {
		        lastPos = [ lastPos[0], command.values[0] ];
		        ctx.lineTo(lastPos[0], lastPos[1]);
		    } else if (command.marker === 'C') {
		        pointOne = [ command.values[0],
		                     command.values[1] ];
		        pointTwo = [ command.values[2],
		                     command.values[3] ];
		        lastPos  = [ command.values[4],
		                     command.values[5] ];
		        ctx.bezierCurveTo(
		                pointOne[0], pointOne[1],
		                pointTwo[0], pointTwo[1],
		                lastPos[0], lastPos[1]);
		    }
	    });
	}

	// Draws and SVG path
	drawSvgPath(ctx, SVGpath) {
		this.drawSvgPathFromCommandList(ctx, this.svgPathToCommands(SVGpath));
	}
	
	// Sets standard node drawing properties
	setStrokeDrawingProps(context, s, d) {
		var rt = 0;
	  	if(!(typeof d["fcol"] === 'undefined')) {
	  		rt += 1;
		  	context.fillStyle = d.fcol;		// sets the color
	  	}
	  	if(!(typeof d["scol"] === 'undefined')) {
	  		rt += 2;
		  	context.strokeStyle = d.scol;		// sets the color
			if(typeof d["swidth"] === 'undefined') {
				context.lineWidth = 1;
			} else {
				context.lineWidth = d.swidth;
			}
		  	if(!(typeof d["sDash"] === 'undefined')) {
			  	context.setLineDash(d.sDash);		// sets the dash style
		  	}
	  	}
	  	
	  	return rt;
	}

	// draws a node
	drawNode(context, s, d) {
		var wd;
		wd = ExtraGraph.setStrokeDrawingProps(context, s, d);
		  
		switch(d.sh) {
		  case "s":	// square. sz: side
	  		if(wd&1)
			  	context.fillRect(-d.sz/2, -d.sz/2, d.sz, d.sz); // draw the node
	  		if(wd&2)
			  	context.strokeRect(-d.sz/2, -d.sz/2, d.sz, d.sz); // draw the node

			// changes the style if the object is selected
			if(s.selected) {
				context.lineWidth = 3;
			  	context.strokeStyle = "#997777";		// sets the color
			  	context.strokeRect(-d.sz/2, -d.sz/2, d.sz, d.sz); // draw the node
			}
			if(s.hovering) {
				context.lineWidth = 2;
			  	context.strokeStyle = "#ff0000";		// sets the color
			  	context.strokeRect(-d.sz*0.6, -d.sz*0.6, d.sz*1.2, d.sz*1.2); // draw the node
			}
			break;
		  case "c": // circle. sz: radius
		  	context.beginPath();
      		context.arc(0, 0, d.sz/2, 0, 2 * Math.PI, false);
	  		if(wd&1)
			  	context.fill(); // draw the node
	  		if(wd&2)
			  	context.stroke(); // draw the node
			// changes the style if the object is selected
			if(s.selected) {
				context.lineWidth = 3;
			  	context.strokeStyle = "#997777";		// sets the color
			  	context.stroke(); // draw the node
			}
			if(s.hovering) {
			  	context.beginPath();
	      		context.arc(0, 0, d.sz*0.6, 0, 2 * Math.PI, false);
				context.lineWidth = 2;
			  	context.strokeStyle = "#ff0000";		// sets the color
			  	context.stroke(); // draw the node
			}
			break;
		  case "r": // rectangle.  w, h: sides
	  		if(wd&1)
			  	context.fillRect(-d.w/2, -d.h/2, d.w, d.h); // draw the node
	  		if(wd&2)
			  	context.strokeRect(-d.w/2, -d.h/2, d.w, d.h); // draw the node

			// changes the style if the object is selected
			if(s.selected) {
				context.lineWidth = 3;
			  	context.strokeStyle = "#997777";		// sets the color
			  	context.strokeRect(-d.w/2, -d.h/2, d.w, d.h); // draw the node
			}
			if(s.hovering) {
				context.lineWidth = 2;
			  	context.strokeStyle = "#ff0000";		// sets the color
			  	context.strokeRect(-d.w*0.6, -d.h*0.6, d.w*1.2, d.h*1.2); // draw the node
			}
			break;
		  case "e": // ellipse. w, h: radiuses 
		  	context.beginPath();
		    var magic = 0.551784;
			var rlx = d.w/2,
		        tby = d.h/2;
		    var xmagic = magic * d.w / 2;
		    var ymagic = magic * d.h / 2;
		    context.moveTo(0, -tby);
		    context.bezierCurveTo(xmagic, -tby, rlx, -ymagic, rlx, 0);
		    context.bezierCurveTo(rlx, ymagic, xmagic, tby, 0, tby);
		    context.bezierCurveTo(-xmagic, tby, -rlx, ymagic, -rlx, 0);
		    context.bezierCurveTo(-rlx, -ymagic, -xmagic, -tby, 0, -tby);
		    
		    if(wd&1)
			  	context.fill(); // draw the node
	  		if(wd&2)
			  	context.stroke(); // draw the node
			// changes the style if the object is selected
			if(s.selected) {
				context.lineWidth = 3;
			  	context.strokeStyle = "#997777";		// sets the color
			  	context.stroke(); // draw the node
			}
			if(s.hovering) {
				rlx = d.w * 0.6;
			    tby = d.h * 0.6;
			    xmagic = magic * d.w * 0.6;
			    ymagic = magic * d.h * 0.6;
			    context.moveTo(0, -tby);
			    context.bezierCurveTo(xmagic, -tby, rlx, -ymagic, rlx, 0);
			    context.bezierCurveTo(rlx, ymagic, xmagic, tby, 0, tby);
			    context.bezierCurveTo(-xmagic, tby, -rlx, ymagic, -rlx, 0);
			    context.bezierCurveTo(-rlx, -ymagic, -xmagic, -tby, 0, -tby);
				context.lineWidth = 2;
			  	context.strokeStyle = "#ff0000";		// sets the color
			  	context.stroke(); // draw the node
			}
			break;
		  case "p": // SVG path. d: the SVG path definition
		  	this.drawSvgPath(context, d.d);
	  		if(wd&1)
			  	context.fill(); // draw the node
	  		if(wd&2)
			  	context.stroke(); // draw the node
			// changes the style if the object is selected
			if(s.selected) {
				context.lineWidth = 3;
			  	context.strokeStyle = "#997777";		// sets the color
			  	context.stroke(); // draw the node
			}
			if(s.hovering) {
				context.lineWidth = 2;
			  	context.strokeStyle = "#ff0000";		// sets the color
			  	context.stroke(); // draw the node
			}
			break;
		  case "g": // Group of elements. d: array with the individual elements.
					//                       each element has extra properties x, y for position
		  	d.d.forEach(function(cmd) {
		  		context.translate(cmd.x, cmd.y);
		  		ExtraGraph.drawNode(context, s, cmd);
		  		context.translate(-cmd.x, -cmd.y);
		  	});
			break;
		  case "t": // Text. w, h: size of the test. d: font. txt: string to print
		  	var h = d.h;
		  	var ft = d.sz + "px " + d.font;
		  	var w = d.w;
		  	context.font = ft;
	  		if(wd&1)
			  	context.fillText(d.txt, w, h); // draw the node
	  		if(wd&2)
			  	context.strokeText(d.txt, w, h); // draw the node
			// changes the style if the object is selected
			if(s.selected) {
				context.lineWidth = 3;
			  	context.strokeStyle = "#997777";		// sets the color
			  	context.strokeText(d.txt, w, h); // draw the node
			}
			if(s.hovering) {
				context.lineWidth = 2;
			  	context.strokeStyle = "#ff0000";		// sets the color
			  	context.strokeText(d.txt, w, h); // draw the node
			}
			break;
		  case "i": // Image. img: url of the image
		  	// load the image the first time the procedure is callled
			if(typeof d.img === 'string') {
				d.img = this.loadImage(this.tool, d.img);
			}
			// draws an empty symbol if the image has not been loaded yet	
			if(d.img.loaded) {
				context.drawImage(d.img, -d.w/2, -d.h/2, d.w, d.h);
			} else {
				context.lineWidth = 1;
			  	context.strokeStyle = "#000000";		// sets the color
			  	context.strokeRect(-d.w/2, -d.h/2, d.w, d.h); // draw the node
				context.beginPath();
				context.moveTo(-d.w/2, -d.h/2);
				context.lineTo( d.w/2,  d.h/2);
				context.moveTo( d.w/2, -d.h/2);
				context.lineTo(-d.w/2,  d.h/2);
				context.closePath();
				context.stroke();
			}

			// changes the style if the object is selected
			if(s.selected) {
				context.lineWidth = 3;
			  	context.strokeStyle = "#997777";		// sets the color
			  	context.strokeRect(-d.w/2, -d.h/2, d.w, d.h); // draw the node
			}
			if(s.hovering) {
				context.lineWidth = 2;
			  	context.strokeStyle = "#ff0000";		// sets the color
			  	context.strokeRect(-d.w*0.6, -d.h*0.6, d.w*1.2, d.h*1.2); // draw the node
			}
			break;
		}
	}

	// draws an arrow
	drawArrow(context, x1, y1, x2, y2, s, d) {
		if(!(typeof d === 'undefined')) {			
			var ang = -Math.atan2(x2 - x1, y2 - y1);
//console.log("Drawing arrow: " + x2 + " " + y2 + " " + ang + " " + JSON.stringify(d) + " ");
//sconsole.trace();
			context.save();
			ExtraGraph.setStrokeDrawingProps(context, s, d);
			context.translate(x2, y2);
			context.rotate(ang);
			ExtraGraph.drawNode(context, s, d);
			context.restore();
		}
	}

	// clears the screen
	clear() {
//	  this.context.save();
//	  this.context.setTransform(1,0,0,1,0,0);
	  this.context.clearRect(0, 0, this.canvas.width, this.canvas.height);
//	  this.context.restore();
	}
	
	//draw a rectangular manipulator
	drawManipulatorRect(tx, ty, rot, sx, sy, el, manip, draw) {
		var context = this.context; // quick access to the context
		var i, j;
		var sz, mcx, mcy;
		context.translate(tx, ty);
		context.rotate(rot);

		var lx = sx + 7;
		var ly = sy + 7;

	  	context.beginPath();
	  	context.strokeRect(-lx, -ly, lx * 2, ly * 2); // draw the manipulator box
	  	context.moveTo(0.5, -ly);
	  	context.lineTo(0.5, -ly-10);
	  	context.stroke();
	  	for(i = -1; i <= 1; i++) {
	  		for(j = -1; j <= 1; j++) {
	  			if((i != 0) || (j != 0)) {
	  				sz = 2;
	  			} else {
	  				sz = 3;
	  			}
	  			mcx = i * lx;
	  			mcy = j * ly;
	  			if(draw.hoveringManipNum == j*3+i+4) {	// for hovering
	  				context.save();
					context.fillStyle   = "#ff0000";
	  			}
		  		context.fillRect(  mcx -sz, mcy -sz,  2*sz+1, 2*sz+1);
	  			if(draw.hoveringManipNum == j*3+i+4) {	// for hovering
	  				context.restore();
	  			}
				context.strokeRect(mcx -sz, mcy -sz,  2*sz+1, 2*sz+1);
				draw.addManipulator(tx + mcx * Math.cos(rot) - mcy * Math.sin(rot), 
									ty + mcy * Math.cos(rot) + mcx * Math.sin(rot),
									sz+1, j*3+i+4, el, manip);
	  		}
	  	}
 		if(draw.hoveringManipNum == 9) {	// for hovering
 			context.save();
			context.fillStyle   = "#ff0000";
 		}
  		context.fillRect(  -2, -ly -10,  5, 5);
		if(draw.hoveringManipNum == 9) {	// for hovering
			context.restore();
		}
		context.strokeRect(-2, -ly -10,  5, 5);
		draw.addManipulator(tx + (ly + 8) * Math.sin(rot), 
							ty - (ly + 8) * Math.cos(rot),
							3, 9, el, manip);
	}


	// generic linear interpolation function
	interp(x1, x, x2, y1, y2) {
		if(x1 != x2) {
			return (y1 + (y2 - y1) * (x - x1) / (x2 - x1));
		} else {
			return ((y1 + y2) / 2);
		}
	}
	
	// interpolates only inside the interval, translates outside
	exInterp(x1, x, x2, y1, y2) {
		if((x1 == x2) && (x == x1)) {
			return ((y1 + y2) / 2)
		} else if(((x < x1) && (x1 <= x2)) || ((x > x1) && (x1 >= x2))) {
			return (x + y1 - x1);
		} else if(((x < x2) && (x1 < x2)) || ((x > x2) && (x1 > x2))) {
			return this.interp(x1, x, x2, y1, y2);
		} else {
			return (x + y2 - x2);
		}
		
	}
	
	// extends the current bounding box to include the given point
	extendBbox(x, y, bbox) {
		if(x < bbox.x) bbox.x = x;
		if(x > bbox.w) bbox.w = x;
		if(y < bbox.y) bbox.y = y;
		if(y > bbox.h) bbox.h = y;
	}
	
	// proportionally expands an edge
	expandEdge(src, last, points, x1n, y1n, x2n, y2n) {
		// src -> initial array of the edges
		// points -> new array that describes the edge
		// x1n, y1n, x2n, y2n -> new extremities of the edge
		var j;
		var x1o =  src[0].x1;
		var y1o =  src[0].y1;
		var x2o =  src[last].x2;
		var y2o =  src[last].y2;
		
		var eEx1, eEy1, eExc1, eEyc1, eEx2, eEy2, eExc2, eEyc2;
		for(j = 0; j <= last; j++) {
			eEx1  = ExtraGraph.exInterp(x1o,src[j].x1 ,x2o,x1n,x2n);
			eExc1 = ExtraGraph.exInterp(x1o,src[j].xc1,x2o,x1n,x2n);
			eExc2 = ExtraGraph.exInterp(x1o,src[j].xc2,x2o,x1n,x2n);
			eEx2  = ExtraGraph.exInterp(x1o,src[j].x2 ,x2o,x1n,x2n);
			eEy1  = ExtraGraph.exInterp(y1o,src[j].y1 ,y2o,y1n,y2n);
			eEyc1 = ExtraGraph.exInterp(y1o,src[j].yc1,y2o,y1n,y2n);
			eEyc2 = ExtraGraph.exInterp(y1o,src[j].yc2,y2o,y1n,y2n);
			eEy2  = ExtraGraph.exInterp(y1o,src[j].y2 ,y2o,y1n,y2n);
			
			if(((eEx1 != x1n) || (eEy1 != y1n)) && (j == 0)) {
				eEx1 = x1n;
				eEy1 = y1n;
				eExc1 = src[j].xc1+x1n-x1o;
				eEyc1 = src[j].yc1+y1n-y1o;
			}
			if(((eEx2 != x2n) || (eEy2 != y2n)) && (j == last)) {
				eEx2 = x2n;
				eEy2 = y2n;
				eExc2 = src[j].xc2+x2n-x2o;
				eEyc2 = src[j].yc2+y2n-y2o;
			}
			
			points[j] = {path: [j], more:[
				{path: ["x1"],  val: eEx1},
				{path: ["y1"],  val: eEy1},
				{path: ["xc1"], val: eExc1},
				{path: ["yc1"], val: eEyc1},
				{path: ["x2"],  val: eEx2},
				{path: ["y2"],  val: eEy2},
				{path: ["xc2"], val: eExc2},
				{path: ["yc2"], val: eEyc2}
			]};
		}
	}
	
	// transform a coordinate. Returns an object with the transformed coordinated in .x and .y
	transformPoint(x, y, manType, transf) {
		var out = {x: x, y: y};
		out.x -= transf.cx;					// first centers the transform
		out.y -= transf.cy;
		if(manType != 2) {		// only if position is involved
			var px = out.x;
			var py = out.y;
			out.x = Math.cos(transf.crot) * px + Math.sin(transf.crot) * py;
			out.y = Math.cos(transf.crot) * py - Math.sin(transf.crot) * px;
	
			out.x *= transf.d_lx;
			out.y *= transf.d_ly;
	
			var srot = transf.crot + transf.d_rot;
			px = out.x;
			py = out.y;
			out.x = Math.cos(srot) * px - Math.sin(srot) * py;
			out.y = Math.cos(srot) * py + Math.sin(srot) * px;
		}
		out.x += transf.d_tx + transf.cx;		// applys the transforms: translation
		out.y += transf.d_ty + transf.cy;

		return out;
	}	

	// Timer functions
	timerHandler() {
		ExtraGraph.currentTimer = "none";
		ExtraGraph.sendTimeout();
//		this.DN4.setState(this.DN4.guiState.timeout());
	}
	// starts the timer to fire after a given timeout
	startTimer(timeout) {
		if(this.currentTimer != "none") {
			clearTimeout(this.currentTimer);
		}
		this.currentTimer = setTimeout(this.timerHandler, timeout);
	}
	// cancels the timer
	stopTimer() {
		if(this.currentTimer != "none") {
			clearTimeout(this.currentTimer);
			this.currentTimer = "none";
		}
	}
	// forces the timeout
	sendTimeout() {
		this.tool.sendTimeout();
	}
	
	// handles shift, zoom and pan
	shift(event) {
		return (event.shiftKey || (this.tool.menu.currentZoomPanShift == "mi_shift"));
	}
	ctrl(event) {
		return (event.ctrlKey || (this.tool.menu.currentZoomPanShift == "mi_zoom"));
	}
	alt(event) {
		return (event.altKey || (this.tool.menu.currentZoomPanShift == "mi_pan"));
	}
	
	loadImage(tool, url) {
		var img = new Image();
		img.loaded = false;
		img.tool = tool;
		img.addEventListener('load', function() {
		  // execute drawImage statements here
		  this.loaded = true;
		  this.tool.controls.draw();
		}, false);
		img.src = url;
		return img;
	}
	
	snap(val) {
		return this.tool.menu.items.mi_snap.selected ? 
			   Math.round(val / this.tool.draw.gridSize) * this.tool.draw.gridSize :
			   val;
	}

	snapAngle(x,y) {
		var ang = Math.atan2(x, y);
		return this.tool.menu.items.mi_snap.selected ? 
			   Math.round(ang / Math.PI * 12) * Math.PI / 12 :
			   ang;
	}
	snapAngle1(ang) {
		return this.tool.menu.items.mi_snap.selected ? 
			   Math.round(ang / Math.PI * 12) * Math.PI / 12 :
			   ang;
	}
	snapScale(val) {
		var scale = Math.pow(10, Math.floor(Math.log(val/2) / Math.log(10)));
		return this.tool.menu.items.mi_snap.selected ? 
			   Math.round(val / scale) * scale : val;
	}
}

// defines the globa variable for the ExtraGraph library
var ExtraGraph = new DN4_ExtraGraph();
