"use strict"	// temporary - for a better debugging

// this function encapsulates the history of commands - used for UNDO / REDO commands
class DN4_history {
	constructor(tool) {
    // all the "closure" variables
		this.DN4 = tool;			// the variable to access all the other components of the gui
		
		this.undoRedoStack = [];		// stack for the UNDO and REDO of commands
										// the stack is composed of commands in the form:
										// {name, undo, redo} where:
										// name: a string that identify the command
										// undo: a command to undo the action (command syntax below)
										// redo: a command to redo the action (command syntax below)
		
		this.currUndo = false;		// current undo command
		this.currRedo = false;		// current redo command
		this.stackSize = 0;			// number of elements in the undo/redo stack
		this.currPos = 0;			// current command in the undo/redo chain		
	}
	// puts the entry of the command composed by currUndo and currRedo in the history.
	// The name of the command is passed
	putInHistory(n) {
		if(!this.currUndo) {
		  if(this.currRedo) {
			alert("Error in Undo/Redo handling: missing UNDO" 
			+this.currUndo + " " + this.currRedo);
		  }
		} else {
		  if(this.currRedo) {
			this.undoRedoStack[this.currPos] = {name: n, redo: this.currRedo, undo: this.currUndo};
			this.currPos++;
			this.stackSize = this.currPos;
		  }
		}
		this.currRedo = false;
		this.currUndo = false;
	}
	
	canUndo() {return (this.currPos > 0);}
	canRedo() {return (this.currPos < this.stackSize);}
	
	makeStackForDebugString() {
		var outStr;
		var i;
		
		outStr =  "Stack size:      " + this.stackSize + "\n";
		outStr += "Curr. entry Id.: " + this.currPos + "\n";
		outStr += "Max. stack size: " + this.undoRedoStack.length + "\n";
		for(i = 0; i < this.stackSize; i++) {
			if(i == this.currPos-1) {
				outStr += ">>> ";
			} else {
				outStr += "    ";
			}
			outStr += this.undoRedoStack[i].name + "\n";
		}
		outStr += "can UNDO: " + this.canUndo() +", can REDO: " + this.canRedo();
		outStr += "\ngraphFirstFree: " + graphFirstFree + ", graph.lenght: " + graph.length;

		return outStr;
	}
	printStackForDebug() {
		alert(this.makeStackForDebugString());
	}
	//// Format for the undo/redo stack entry:
	// [                -- array of commands
	// {cmd: string,	-- identify the command 
	// the rest of the object is command dependent
	// for cmd = "p" -> modify element graphical property
	//      el: object,	  		     -- object to modify
	//      path: [array of string], -- path to the property to modify
	//      --val:   object          -- value of the property, or
	//      --more: {path, val or sub ...} -- recursive path to set more values
	//                                     -- sharing the same prefix
	//      --add:   index           -- adds an element to the array at the given position
	//      --rem:   index           -- removes the element of the array at the given position
	//      /*NOTE*/ Exactly one of the previous four alternatives must be present
	//      }
	// ]
	// for cmd = "n" -> nop ---- NO extra field
	// for cmd = "s" -> select element
	//      add: [array of elements] -- selects the elements in the array
	//   or add: "all" -- selects all elements
	//      rem: [array of elements] -- remove the elements from the selection
	//   or rem: "all" -- removes all elements from the selection
	// for cmd = "g" -> set group manipulator common properties
	//      tx, ty, lx, ly, rot, cx, cy -- the properties of the group manipulator
	// for cmd = "m" -> add or remove elements
	//      add: [array of elements] -- adds the elements in the array
	//           each element has the following syntax:
	//           {el, t, s} where:
	//               el:  the storage area that will hold the new element
	//               t:   the index of the type of the new element
	//               s:   the data associated to the new element	
	//      rem: [array of elements] -- remove all the elements from the model
	//      ref: [array of elements] -- patch the ref. counter of all the elements
	//      ff:  the first free element at the moment of creation [dirty - to improve]
	// for cmd = "r" -> renames the element
	//      el: object,	  -- object to modify
	//      n: 			  -- new name of the object
	//      o: 			  -- old name of the object
	// for cmd = "v" -> modify element model property
	//      el: object,	  		     -- object to modify
	//      v:  object               -- values of the properties to set
	
	// very tricky: to account for the way in which references works in Javascript!
	recursiveSet(iv, e) {
		var i;
		var v = iv;
		if('more' in e) {
			for(i = 0; i < e.path.length; i++) {
				v = v[e.path[i]];
			}
			for(i = 0; i < e.more.length; i++) {
				this.recursiveSet(v, e.more[i]);
			}
		} else if('val' in e) {
			for(i = 0; i < e.path.length-1; i++) {
				v = v[e.path[i]];
			}
			v[e.path[i]] = e.val;
		} else if('add' in e) {
			for(i = 0; i < e.path.length-1; i++) {
				v = v[e.path[i]];
			}
			v = v[e.path[i]];
			v.splice(e.add, 0, {});
		} else if('rem' in e) {
			for(i = 0; i < e.path.length-1; i++) {
				v = v[e.path[i]];
			}
			v = v[e.path[i]];
			v.splice(e.rem, 1);
		}
	}

	modelRecursiveSet(iv, e) {
		var i;
		var v = iv;
		
//alert("H\n\n" + JSON.stringify(iv) + "\n\n" + JSON.stringify(e));
		
		for (const key in e) {
			if (e.hasOwnProperty(key)) {
				var ev = e[key];
//alert(key + " ... " + JSON.stringify(ev) + " .old. " + JSON.stringify(v[key]));
				if (typeof ev === 'object' && !Array.isArray(ev) && ev !== null) {
					this.modelRecursiveSet(v[key], ev)
				} else {
					v[key] = ev;
				}
			}
		}
	}
	
	execute(entry) {
		var i, j, e;
		var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
		var model = this.DN4.model;	// speed up accessing the model methods
		var willRedraw = false;

		for(i = 0; i < entry.length; i++) {
			e = entry[i];
			switch(e.cmd) {
			  case "p":
			  	// sets the properties
			    this.recursiveSet(e.el.s, e);
				// adjust the R-tree
				var s = e.el.s;
				var d = elementTypes[e.el.t].d;
				var t = elementTypes[e.el.t].t;
				draw.sceneTree.remove(e.el.bbox, e.el);
				e.el.bbox = this.DN4.metaTypeElements[t].getBbox(s, d);
				draw.sceneTree.insert(e.el.bbox, e.el);
				break;
			  case "s":
			  	// selects or deselects elements
			  	if("add" in e) {	// elements to add to the selection
			  		if(e.add == "all") { // adds all elements to the selection
alert("select all elements - NOT YET IMPLEMENTED");
			  		} else {	// selection of elements from a list
			  			for(j = 0; j < e.add.length; j++) {
							draw.addToCurrentSelection(e.add[j]);	// adds it to the selection
			  			}
			  		}
			  	}
			  	if("rem" in e) {	// elements to remove from the selection
			  		if(e.rem == "all") { // remove all elements from the selection
						draw.clearCurrentSelection();
			  		} else {	// deselection of elements from a list
			  			for(j = 0; j < e.rem.length; j++) {
							draw.removeFromCurrentSelection(e.rem[j]);
			  			}
			  		}
			  	}
				draw.initGroupManipulators();	// check and updates group manipulators (if needed)
				willRedraw = true;
			  	break;
			  case "g":
			    // set group manipulator common properties
			    draw.restoreGroupManipulators(e);
			    break;
			  case "m":
			  	// adds or removes elements from the model
			  	if("add" in e) {	// elements to add to the model
		  			for(j = 0; j < e.add.length; j++) {
		  				// adds the element to the model
						model.putElement(e.add[j].el, e.add[j].t, e.add[j].s);
		  			}
			  	}
			  	if("rem" in e) {	// elements to remove from the model
//alert("start removing " + e.rem.length + " element(s)");
		  			for(j = 0; j < e.rem.length; j++) {
		  				model.removeElement(e.rem[j]);
		  			}
//alert("end removing");
			  	}
			  	if("ref" in e) {	// elements to update to reference count
		  			for(j = 0; j < e.ref.length; j++) {
		  				// updates the reference count of the element
						e.ref[j].el.r += e.ref[j].dr;
		  			}
			  	}
/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/
			  	if("ff" in e) {						// to set the first free element
			  		model.setFirstFree(e.ff);
			  	}
/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/
				draw.initGroupManipulators();	// check and updates group manipulators (if needed)
				willRedraw = true;
			  	break;
			  case "n":
			    // NOP - do nothing
			    alert("NOP command issued!");
			    break;
			  case "r":
			  	// renames the object
				e.el.n = e.n;
				usedNames[e.n] = 1;
				usedNames[e.o] = 0;
			    break;
			  case "v":
			  	// changes values of a model property
				this.modelRecursiveSet(e.el.p, e.v);
				break;
			}
		}
		if(willRedraw) {
			draw.drawView();
		}
	}
	
	setCurrUndo(entry) {	// sets the undo for the current commnad
		this.currUndo = entry;
	}
	       	
	setCurrRedo(entry) {	// sets the redo for the current commnad
		this.currRedo = entry;
	}
	
	execWithHistory(entry) {	// execs a command, with history enabled
		this.setCurrRedo(entry);
		this.execute(entry);
	}
	
	doUndo() {
		if(this.canUndo()) {
			this.currPos--;
			this.execute(this.undoRedoStack[this.currPos].undo);
			this.DN4.menu.update();
		}
	}

	doRedo() {
		if(this.canRedo()) {
			this.execute(this.undoRedoStack[this.currPos].redo);
			this.currPos++;
			this.DN4.menu.update();
		}
	}
	addAndExec(n, u, r) {
		this.setCurrUndo(u);
		this.execWithHistory(r);
		this.putInHistory(n);
		this.DN4.menu.update();
	}	       	
};