"use strict"	// temporary - for a better debugging

// this is the main component of the DrawNET 4 GUI
class DrawNET4 {
	constructor() {
		// initializes the components of the GUI
		this.config           = new DN4_config(this);
		this.metaTypeElements = new DN4_metaTypeElements(this);
		this.cmdHistory       = new DN4_history(this);
		this.draw             = new DN4_draw(this);
		this.menu             = new DN4_menu(this);
		this.controls         = new DN4_controls(this);
		this.FSM              = new DN4_FSM(this);
		this.model            = new DN4_model(this);
		
		this.guiState = this.FSM.GS_NORMAL;	// state machine of the GUI interaction
		this.setState = function(newState) {
			if(!(newState === this.guiState)) {
				this.guiState.exit();
				this.guiState = newState;
				this.guiState.enter();
			}
		}
	 }
	sendTimeout() {
		this.setState(this.guiState.timeout());
	}
}

// this function creates the GUI
$(document).ready(function() {
	var DN4 = new DrawNET4();

	// refresh internal, canvas and controls dimensions after resize
	$(window).resize(function() {
		DN4.draw.resize();
		DN4.controls.resize();
	});
 
 	// sets the mouse events to the handler in the current state
 	$("#mainCanvas").mousedown(function(event) {
		DN4.setState(DN4.guiState.mousedown(event));
 	});
 	$("#mainCanvas").mouseup(function(event) {
		DN4.setState(DN4.guiState.mouseup(event));
 	});
	$("#mainCanvas").mousemove(function(event) {
		DN4.setState(DN4.guiState.mousemove(event));
	});
	
 	// sets the mouse events to the handler in the current state
 	$("#ctrlCanvas").mousedown(function(event) {
		DN4.controls.mousedown(event);
 	});
 	$("#ctrlCanvas").mouseup(function(event) {
		DN4.controls.mouseup(event);
 	});
	$("#ctrlCanvas").mousemove(function(event) {
		DN4.controls.mousemove(event);
	});
	$("#ctrlCanvas").mouseout(function() {
		DN4.controls.mouseout();
	});

	// to support keyboard shortcuts
	$(document).keyup(function(event) {
// handling keys with dialogs has become more difficult.
// To be addressed in the future!
/*		var knownKey = true;
		if((event.keyCode == 90) && (event.ctrlKey)) {
		   	// CTRL + Z -----> UNDO
			DN4.cmdHistory.doUndo();
			DN4.draw.drawView();
		} else if((event.keyCode == 89) && (event.ctrlKey)) {
		   	// CTRL + Y -----> REDO
			DN4.cmdHistory.doRedo();
			DN4.draw.drawView();
		} else if((event.keyCode == 72) && (event.ctrlKey)) {
		   	// CTRL + H -----> prints the history
			// // // Now it should show the DebugDialogWindow in model.js DebugModelDialog(1)
		} else if((event.keyCode == 71) && (event.ctrlKey)) {
		   	// CTRL + G -----> prints the entire graph
			alert(graphFirstFree + "\n" + JSON.stringify(graph));
		} else if((event.keyCode == 68) && (event.ctrlKey)) {
		   	// CTRL + D -----> prints the selected elements
			// // // Now it should show the DebugDialogWindow in model.js DebugModelDialog(2)
		} else if((event.keyCode == 8) || (event.keyCode == 46)) {
		   	// Delete or Backspace -----> deletes the selected elements
			if(DN4.draw.currentSelection.length > 0) {
				DN4.model.removeSelectedElements();
			}
		} else {
			knownKey = false;
		}
		if(knownKey) {
			event.preventDefault();
		}*/
	});

	// disable context menu (and CTRL click on macs)
	$("#mainCanvas, #ctrlCanvas, a").contextmenu(function(e) {
                   e.preventDefault();
            });
            
    // initializes the graphic window
	DN4.draw.init();
    // initializes the control window
	DN4.controls.init();
	
	// packs the window and the controls against the main menu
	DN4.draw.resize();
	DN4.controls.resize();
	
	// installs the main menu handler
	$("a").click(function(event) {
		DN4.menu.click(jQuery(this).attr("id"), event);
		return false;
	}).mouseover(function(event) {
		DN4.menu.mouseover(jQuery(this).attr("id"), event);
	}).mouseout(function(event) {
		DN4.menu.mouseout(jQuery(this).attr("id"), event);
	});
	// initializes menus
	DN4.menu.update();
	
	// prevent scrollbars to appear
	$('html, body').css({
	    'overflow': 'hidden',
	    'height': '100%'
	});
});
