"use strict"	// temporary - for a better debugging

var dialogChangeLog = [];	// store values to update when Apply or Ok are pressed.

// this function encapsulates the model creation and manipulation
class DN4_model {
	constructor(tool) {
	    // all the "closure" variables
		this.DN4 = tool;			// the variable to access all the other components of the gui
	}
	// allocates an element and returns its location
	getFree() {
		// gets the first free element
		var id = graphFirstFree;
//alert(id + " next: " + graph[id].s.next);
		// removes it from the empty list
		graphFirstFree = graph[id].s.next;
		
		// if there are no more free elements, it adds one
		if(graphFirstFree == -1) {
			graphFirstFree = graph.length;
			graph[graphFirstFree] = {id: graphFirstFree, s: {next: -1},
									 t: -1, r:1, bbox: {x: 0, y: 0, w: 0, h: 0}, n: "", p:{}};
		}
		
		return graph[id];
	}

	// inserts an element into an empty position (and update the rtree)
	putElement(el, t, s) {
		if(el.t == -1) {		// can only put an element into an empty place
			var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
			var d = elementTypes[t].d;
			var mt = elementTypes[t].t;
			el.t = t;
			el.s = s;
			el.r = 0;
//alert(el.id + " " + t + " " + JSON.stringify(s));
			el.bbox = this.DN4.metaTypeElements[mt].getBbox(s, d);
			draw.sceneTree.insert(el.bbox, el);
			el.n = elementTypes[t].ft + namesFirstFree;
			namesFirstFree++;
			el.p = this.InitNewElementProperties(el.t);
			usedNames[el.n] = 1;
		} else {
			alert("Error: trying to insert and element into a full position. \n" +
				  "el: " + el.id + ", t: " + el.t + ", r: " + el.r);
		}
	}
/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/			
			// This solution is not nice: after setting an empty element,
			// the user must also set the first free.
			// In this way, undo can freely removing one element
			// and redoing will reset the pointer to the first element
			// that was available before clearing everyhing.
			// // // FOR THE MOMENT IT STAYS LIKE THIS
			//
			// STILL HAVE TO SEE IF THIS WORKS ALSO FOR ELEMENT DELETION
			//
	setFirstFree(ff) {
			graphFirstFree = ff;
	}
/**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**//**/			
	
	// removes an elment from the model (and from the rtree)
	removeElement(oel) {
		if(oel.t == -1) {
			alert("Error: trying to delete an element from a free position. \n" +
				  "el: " + oel.id + ", t: " + oel.t + ", r: " + oel.r);
		} else {
			if(oel.r > 0) {
				alert("Error: trying to delete an element that has references. \n" +
				  "el: " + oel.id + ", t: " + oel.t + ", r: " + oel.r);
			} else {
				// removes the element from the r-tree
				this.DN4.draw.sceneTree.remove(oel.bbox, oel);
				// rests its content
				oel.t = -1;
				oel.r = 1;
				oel.s = {next: graphFirstFree};
				oel.bbox = {x: 0, y: 0, w: 0, h: 0};
				oel.n = "";
				oel.p = {};
				graphFirstFree = oel.id;
			}
		}
	}	
	
	
	
	
	
	
	
	
	
	
	
	
			////////////////////////////////
			//// MODEL HELPER FUNCTIONS ////
			////////////////////////////////
			// These functions sempifies the interaction between the tool and the model.
			// Maybe this is not the right place to put them and in the future they will be
			// moved in another component	



	
	// constructs the history entry to delete an element (a node, or an edge in a group)
	removeFullElementCmd(el, redo, undo) {
		// el is the element to remove
		// redo is the array with the commands to execute the deletion
		// undo is the array with the commands to undo the deletion

		var i;

		// UNDO
		i = undo.length;
		undo[i] = {cmd: "m", add: [{el: el, t: el.t, s: el.s}]};
		i++;
		undo[i] = {cmd: "m", ref: [{el: el, dr: el.r}]};

		// COMMAND
		i = redo.length;
		redo[i] = {cmd: "m", ref: [{el: el, dr: -el.r}]};
		i++;
		redo[i] = {cmd: "m", rem: [el]};		
	}
	
	// constructs the history entry to delete a single edge from the model
	removeSingleEdgeCmd(el, ext, redo, undo) {
		// el is the element to remove
		// ext determines which end points will not be deleted:
		//             "b" - both: only the edge is removed, so the nodes on both sides must be updated
		//             "i" - in: the node in the "to" remains and must be updated
		//             "o" - out: the node in the "from" remains and must be updated
		// redo is the array with the commands to execute the deletion
		// undo is the array with the commands to undo the deletion

		var i, j;

		// UNDO
		i = undo.length;
		undo[i] = {cmd: "m", add: [{el: el, t: el.t, s: el.s}]};
		i++;
		undo[i] = {cmd: "m", ref: [{el: el, dr: 2}]};
		// COMMAND
		i = redo.length;
		redo[i] = {cmd: "m", ref: [{el: el, dr: -2}]};
		i++;
		redo[i] = {cmd: "m", rem: [el]};		

		// corrects the input and output nodes
		if((ext == "b") || (ext == "i")) {	// the to of the arc is in an input node that is deleted
											// we must correct the out of the node in the from
			var fromEl = graph[el.s.from];
			for(j = 0; j < fromEl.s.out.length; j++) {
				if(fromEl.s.out[j] == el.id) {	// found the reference in the node
					break;
				}
			}
			if(j == fromEl.s.out.length) {
				alert("Error! A node in the graph does not refer to an arc that starts from it");
			} else {
				// UNDO
				i = undo.length;
				undo[i] = {cmd: "p", el: fromEl, path: ["out"], add: j};
				i++;
				undo[i] = {cmd: "p", el: fromEl, path: ["out", j], val: el.id};
				i++;
				undo[i] = {cmd: "m", ref: [{el: fromEl, dr: 1}]};
				// COMMAND
				i = redo.length;
				redo[i] = {cmd: "m", ref: [{el: fromEl, dr: -1}]};
				i++;
				redo[i] = {cmd: "p", el: fromEl, path: ["out"], rem: j};		
			}
		}

		if((ext == "b") || (ext == "o")) {	// the from of the arc is in an output node that is deleted
											// we must correct the in of the node in the to
			var toEl = graph[el.s.to];
			for(j = 0; j < toEl.s.in.length; j++) {
				if(toEl.s.in[j] == el.id) {	// found the reference in the node
					break;
				}
			}
			if(j == toEl.s.in.length) {
				alert("Error! A node in the graph does not refer to an arc that ends on it");
			} else {
				// UNDO
				i = undo.length;
				undo[i] = {cmd: "p", el: toEl, path: ["in"], add: toEl.s.in.length-1};
				i++;
				undo[i] = {cmd: "p", el: toEl, path: ["in", toEl.s.in.length-1], val: el.id};
				i++;
				undo[i] = {cmd: "m", ref: [{el: toEl, dr: 1}]};
				// COMMAND
				i = redo.length;
				redo[i] = {cmd: "m", ref: [{el: toEl, dr: -1}]};
				i++;
				redo[i] = {cmd: "p", el: toEl, path: ["in"], rem: j};		
			}
		}
	}
	
	// removes the currently selected elements from the model
	removeSelectedElements() {
		var draw = this.DN4.draw;	// speed up accessing the draw object of the GUI
		var model = this.DN4.model;	// speed up accessing the model methods
		var i, j;
		var redo = [];
		var undo = [];
		var cmdName;

		// in the command, as first step, clear the selection
		redo[0] = {cmd: "s", rem: "all"};

		if(draw.justEdgesSelected) {
			cmdName = "Delete Edges";
//			alert("Deleting " + draw.currentSelection.length + " edges");
			for(i = 0; i < draw.currentSelection.length; i++) {
				model.removeSingleEdgeCmd(draw.currentSelection[i], "b", redo, undo);
			}
		} else {
			cmdName = "Delete Elements";
//			var nNodes, nFullEdges, nPartialEdges;
//			nNodes = nFullEdges = nPartialEdges = 0;
			for(i = 0; i < draw.currentSelection.length; i++) {
				if(elementTypes[draw.currentSelection[i].t].t == 'e') {
//					nFullEdges++;
					model.removeFullElementCmd(draw.currentSelection[i], redo, undo);
				} else {
					model.removeFullElementCmd(draw.currentSelection[i], redo, undo);
//					nNodes++;
					for(j = 0; j < draw.currentSelection[i].s.in.length; j++) {
						if(!graph[draw.currentSelection[i].s.in[j]].s.selected) {
							model.removeSingleEdgeCmd(graph[draw.currentSelection[i].s.in[j]],
												   "i", redo, undo);
//							nPartialEdges++;
						}
					}
					for(j = 0; j < draw.currentSelection[i].s.out.length; j++) {
						if(!graph[draw.currentSelection[i].s.out[j]].s.selected) {
							model.removeSingleEdgeCmd(graph[draw.currentSelection[i].s.out[j]],
												  "o", redo, undo);
//							nPartialEdges++;
						}
					}
				}
			}
//			alert("Deleting " + nNodes + " nodes, " + nFullEdges + " full edges and " +
//				  nPartialEdges + " partial edges");
		}
		// in the undo, it must also restore the pointer to the first free element
		undo[undo.length] = {cmd: "m", ff: graphFirstFree};
		// in the undo, as last step, restores the selection
		undo[undo.length] = {cmd: "s", rem: "all"}; 
		undo[undo.length] = {cmd: "s", add: draw.currentSelectionToArray()};
		var gmp = {cmd: "g"};	// to resotre group manipulators positions
		draw.saveGroupManipulators(gmp);
		undo[undo.length] = gmp;

		// executes the deletion command
		this.DN4.cmdHistory.addAndExec(cmdName, undo, redo);
	}


	// Save and Debug features
	MakeFromJSONArrayOnePerLine(Pre,A) {
		var i;
		var out = "[\n";
		for(i = 0; i < A.length; i++) {
			out = out + Pre + JSON.stringify(A[i]) + (i < A.length-1 ? "," : "") + "\n";
		}
		out = out + "];\n";
		return out;
	}
	MakeSaveDebugModelString(forSave) {
		var i;
		var pPn = "graph = " + this.MakeFromJSONArrayOnePerLine("   ", graph);
		pPn = pPn + "graphFirstFree = " + graphFirstFree + ";\n";
		pPn = pPn + "namesFirstFree = " + namesFirstFree + ";\n";
		if(!forSave) {
			pPn = pPn + "usedNames = " + JSON.stringify(usedNames) + ";\n";
			pPn = pPn + "\n\n";
			pPn = pPn + "propertyTypes = " + this.MakeFromJSONArrayOnePerLine("   ", propertyTypes);
			pPn = pPn + "ptMap = " + JSON.stringify(ptMap) + ";\n";
			pPn = pPn + "propertyCollection = " + this.MakeFromJSONArrayOnePerLine("   ", propertyCollection);
			pPn = pPn + "pcMap = " + JSON.stringify(pcMap) + ";\n";
			pPn = pPn + "elementTypes = " + this.MakeFromJSONArrayOnePerLine("   ", elementTypes);			
		}
		
		return pPn;
	}
	
	ShowModelInDebugDialog() {
		$("#dialog_items").remove();
		var pPn = "<DIV id=\"dialog_items\"><PRE>";
		pPn = pPn + this.MakeSaveDebugModelString(false);
		pPn = pPn + "</PRE></DIV>"
		$("#dialog").append(pPn);
	}

	ShowHistoryInDebugDialog() {
		$("#dialog_items").remove();
		var pPn = "<DIV id=\"dialog_items\"><PRE>";
		pPn = pPn + this.DN4.cmdHistory.makeStackForDebugString();
		pPn = pPn + "</PRE></DIV>"
		$("#dialog").append(pPn);
	}

	ShowSelectionInDebugDialog() {
		var DN4 = this.DN4;
		$("#dialog_items").remove();
		var pPn = "<DIV id=\"dialog_items\"><PRE>";

		var i;
		if(DN4.draw.currentSelection.length == 0) {
			pPn = pPn + "No element selected";
		} else {
			for(i = 0; i < DN4.draw.currentSelection.length; i++) {
				pPn = pPn + i + ": " + JSON.stringify(DN4.draw.currentSelection[i]) + "\n";
			}
		}

		pPn = pPn + "</PRE></DIV>"
		$("#dialog").append(pPn);
	}

	// Debug Model Dialog
	DebugModelDialog(mode) {
		switch(mode) {
		  case 0:
			this.ShowModelInDebugDialog();
		    break;
		  case 1:
			this.ShowHistoryInDebugDialog();
		    break;
		  case 2:
			this.ShowSelectionInDebugDialog();
		    break;
		  default:
			$("#dialog_items").remove();
			$("#dialog").append("<DIV id=\"dialog_items\">Unkown dialog mode: " + mode + "</DIV");
		    break;
		}
		$("#dialog").data('DN4_model',this);

		$( function() { $( "#dialog" ).dialog({title: "Debug window",
					width: 500,
					height: 450,
					buttons: {
						"Model" : function() {
			var DN4_model = $("#dialog").data('DN4_model');
			DN4_model.ShowModelInDebugDialog();
						},
						"History" : function() {
			var DN4_model = $("#dialog").data('DN4_model');
			DN4_model.ShowHistoryInDebugDialog();
						},
						"Selection" : function() {
			var DN4_model = $("#dialog").data('DN4_model');
			DN4_model.ShowSelectionInDebugDialog();
						},
						"Close" : function(){$(this).dialog('close');}
					}});} );
	}
	
	//// Save model
	SaveModelDialog() {
		$("#dialog_items").remove();
		var pPn = "<DIV id=\"dialog_items\"><PRE>";
		pPn = pPn + this.MakeSaveDebugModelString(true);
		pPn = pPn + "</PRE></DIV>"
		$("#dialog").append(pPn);

		$( function() { $( "#dialog" ).dialog({title: "Save Current model",
					width: 500,
					height: 450,
					buttons: {		
					}});} );
	}
	
	//// Handle properties
	MakeElementInitialPropertyValues(fPs, def) {
		var i;
		var outP = {};
		var useDef = (Object.keys(def).length == fPs.length);
//		alert(fPs + "\n" + useDef + "\n" + JSON.stringify(def));
		for(i = 0; i < fPs.length; i++) {
			var pDef = propertyCollection[fPs[i]];
			var pTy = propertyTypes[pDef.t];
			switch(pTy.t) {
			  case "s": // String
			  case "i": // Integer
			  case "f": // Floating point
			  case "e": // Enum
				outP[pDef.n] = useDef ? def[pDef.n] : pDef.d;
				break;
				break;
			  case "g": // Group
				outP[pDef.n] = this.MakeElementInitialPropertyValues(pTy.r, pDef.d);
				break;
			}
		}
		return outP;
	}	

	InitNewElementProperties(selElT) {
		var selElTdef = elementTypes[selElT];
		var fPs = selElTdef.p;

		return this.MakeElementInitialPropertyValues(fPs, []);
	}
	
	
	
	MakeOnChangeScript(fieldId, checkVal, cthen, celse) {
		var out = " oninput = \"";
		out = out + "if($(&quot;#"+fieldId+"&quot;).val() == &quot;"+checkVal+"&quot;) {";
		out = out + cthen;
		out = out + "} else {";
		out = out + celse;
		return out + "}\" ";
	}
	MakeOnChangeScriptElse(rowId, fieldId, checkVal, celse, bkCol) {
		return this.MakeOnChangeScript(fieldId, checkVal,
					"$(&quot;#"+rowId+"&quot;).data('v_ch',0);"+
					"$(&quot;#"+rowId+"&quot;).css('background-color','" + bkCol + "');",
					celse);
	}
	MakeTextOnChangeScript(rowId, fieldId, checkVal, bkCol) {
		return this.MakeOnChangeScriptElse(rowId, fieldId, checkVal,
					"$(&quot;#"+rowId+"&quot;).data('v_ch',1);"+
					"$(&quot;#"+rowId+"&quot;).css('background-color','#88ff88');", bkCol);
	}
	MakeNameOnChangeScript(rowId, fieldId, checkVal, bkCol) {
		var celse = "if(usedNames[$(&quot;#"+fieldId+"&quot;).val()] == 1) {";
		celse = celse + "$(&quot;#"+rowId+"&quot;).data('v_ch',0);";
		celse = celse + "$(&quot;#"+rowId+"&quot;).css('background-color','#ff8888');";
		celse = celse + "} else {";
		celse = celse + "$(&quot;#"+rowId+"&quot;).data('v_ch',1);";
		celse = celse + "$(&quot;#"+rowId+"&quot;).css('background-color','#88ff88');}";
		return this.MakeOnChangeScriptElse(rowId, fieldId, checkVal,
					celse, bkCol);
	}
	
/*	RevertToPrevious() {		// Revert seems too complex to create, and not worth
								// implementation has been delayed!
								//
		var draw = this.DN4.draw;
		var selEl = draw.currentSelection[0];
		var selElTdef = elementTypes[selEl.t];
		var fPs = selElTdef.p;
		var eName = selEl.n;

		dialogChangeLog = [];
		
		// special code to reset the name
		$("#d_i_name").val(eName);
		$("#rid_d_i_name").css('background-color','')
		$("#rid_d_i_name").data('v_ch',0)
	}*/
	
	GetVariedModelProperties(fPs,  preN, v) {	// fPs  : Array of propeties indices form the PropertyCollection
												// preN : string to add in index names
												// v    : initial values (to build the undo command)
		var i;
		var redo = {};
		var undo = {};
//alert("GVMP\n" + JSON.stringify(fPs) + "\n\n" + JSON.stringify(preN) + "\n\n" + JSON.stringify(v));
		for(i = 0; i < fPs.length; i++) {
			var pDef = propertyCollection[fPs[i]];
			var iName = "d_i_" + preN + i;
			var rName = "rid_d_i_" + preN + i;
			var pTy = propertyTypes[pDef.t];
//alert(JSON.stringify(v));
			var o_v = (pDef.n in v ? v[pDef.n] : pDef.d);

			if(($("#" + rName).data('v_ch') == 1) || (pTy.t == "g")){
				switch(pTy.t) {
				  case "s": // String
				    var n_v = $("#" + iName).val();
					redo[pDef.n] = n_v;
					undo[pDef.n] = o_v;
					break;
				  case "i": // Integer
				    var n_v = $("#" + iName).val();
					redo[pDef.n] = n_v;
					undo[pDef.n] = o_v;
					break;
				  case "f": // Floating point
				    var n_v = $("#" + iName).val();
					redo[pDef.n] = n_v;
					undo[pDef.n] = o_v;
					break;
				  case "e": // Enum
				    var n_v = $("#" + iName).val();
					redo[pDef.n] = n_v;
					undo[pDef.n] = o_v;
					break;
				  case "g": // Group
					var tmp = this.GetVariedModelProperties(pTy.r, preN + "_" + i, o_v);
					redo[pDef.n] = tmp[0];
					undo[pDef.n] = tmp[1];
					break;
				  default:
					break;
				}
			}
		}
		return [redo, undo];
	}	

	MakeRenameCommand(oldName, newName, toEl) {
		var undo = {cmd: "r", el: toEl, n: oldName, o:newName};
		var redo = {cmd: "r", el: toEl, n: newName, o:oldName};
		return [redo, undo];
	}

	gpnames = [
{tt: "x coordinate",              l: "x",   n: "tx",  ri: "rid_d_i_g_x",   fi: "d_i_g_x", s:1},
{tt: "y coordinate",              l: "y",   n: "ty",  ri: "rid_d_i_g_y",   fi: "d_i_g_y", s:1},
{tt: "Rotation in degrees",       l: "rot", n: "rot", ri: "rid_d_i_g_rot", fi: "d_i_g_rot", s:(180.0/Math.PI)},
{tt: "Horizontal scaling factor", l: "hs",  n: "sx",  ri: "rid_d_i_g_sx",  fi: "d_i_g_sx", s:1},
{tt: "Vertical scaling factor",   l: "vs",  n: "sy",  ri: "rid_d_i_g_sy",  fi: "d_i_g_sy", s:1}
	];
	
	UpdateProperties(apply) {
		var draw = this.DN4.draw;
		var selEl = draw.currentSelection[0];
		var selElTdef = elementTypes[selEl.t];
		var fPs = selElTdef.p;
		var eName = selEl.n;

		var moreUndo = [];
		var moreRedo = [];
		var redo = [];
		var undo = [];
		var curCmd = 0;
		var curProp = 0;		
		
		var tmp;
		
		// Name
		var fname = $("#d_i_name").val();
		// special code to set the name
		if($("#rid_d_i_name").data('v_ch') == 1) {
			tmp = this.MakeRenameCommand(eName, fname, selEl);
			redo[curCmd] = tmp[0];
			undo[curCmd] = tmp[1];
			curCmd++;
			// Name has changed
/*			if(apply) {			// Apply seems too complex to create, and not worth
								// implementation has been delayed!
								//
				$("#rid_d_i_name").css('background-color','');
				$("#rid_d_i_name").data('v_ch',0);
				$("#d_i_name").on('input',MakeOnChangeScript(fieldId, checkVal, cthen, celse));
			}*/
		} else {
			if((fname != eName) && (usedNames[fname] == 1)) {
				alert("Name <" + fname + "> already exists. Not changed");
			}
		}


		// Graph properties		
		if(selElTdef.t == "n") {
			var gpnames = this.gpnames;
			for(i = 0; i < gpnames.length; i++) {
				if($("#" + gpnames[i].ri).data('v_ch') == 1) {
					var o_v = selEl.s[gpnames[i].n] * gpnames[i].s;
					var n_v = $("#" + gpnames[i].fi).val() / gpnames[i].s;
					moreRedo[curProp] = {path: [gpnames[i].n], val: n_v};
					moreUndo[curProp] = {path: [gpnames[i].n], val: o_v};
					curProp++;
				}
			}
			
			if(curProp > 0) {
				redo[curCmd] = {cmd: "p", el: selEl, path: [], more: moreRedo};
				undo[curCmd] = {cmd: "p", el: selEl, path: [], more: moreUndo};
				curCmd++;
			}
		}
		
		// Model properties
		tmp = this.GetVariedModelProperties(fPs, "", selEl.p);
		if(Object.keys(tmp[0]).length > 0) {
			redo[curCmd] = {cmd: "v", el: selEl, v: tmp[0]};
			undo[curCmd] = {cmd: "v", el: selEl, v: tmp[1]};
//alert(" redo: " + JSON.stringify(tmp[0]) + "\n\n undo: " + JSON.stringify(tmp[1]));
			curCmd++;
		}

		if(curCmd > 0) {
//			alert(curCmd + "\n\n" + JSON.stringify(redo) + "\n\n" + JSON.stringify(undo));
			this.DN4.cmdHistory.addAndExec("Update Properties", undo, redo);
		}
	}

	MakeElementPropertyTable(fPs, preN, v) {		// fPs  : Array of propeties indices form the PropertyCollection
												// preN : string to add in index names
												// v    : values assigned to the properties
		var i;
		var pPn = "";
		for(i = 0; i < fPs.length; i++) {
			var pDef = propertyCollection[fPs[i]];
			var iName = "d_i_" + preN + i;
			var rName = "rid_d_i_" + preN + i;
			var bkCol = (i % 2 == 0 ? "#ccc" : "#fff");
			pPn = pPn + "<TR title=\"" + pDef.tt + "\" id=\"" + rName + "\" style=\"background-color: " + bkCol + ";\"><TH>" + pDef.l + "</TH><TD ALIGN=\"RIGHT\">";
			var pTy = propertyTypes[pDef.t];
			var p_v = (pDef.n in v ? v[pDef.n] : pDef.d);
			switch(pTy.t) {
			  case "s": // String
//				pPn = pPn + "String : " + iName;
				pPn = pPn + "<INPUT type=\"text\" id=\"" + iName + "\" value=\"" + p_v + "\"" + this.MakeTextOnChangeScript(rName, iName, p_v, bkCol) +">";
				break;
			  case "i": // Integer
//				pPn = pPn + "Integer : " + iName;
				pPn = pPn + "<INPUT type=\"text\" id=\"" + iName + "\" value=\"" + p_v + "\"" + this.MakeTextOnChangeScript(rName, iName, p_v, bkCol) +">";
//alert("<INPUT type=\"text\" id=\"" + iName + "\" value=\"" + pDef.d + "\"" + this.MakeTextOnChangeScript(rName, iName, pDef.d) +">");
				break;
			  case "f": // Floating point
//				pPn = pPn + "Floating point : " + iName;
				pPn = pPn + "<INPUT type=\"text\" id=\"" + iName + "\" value=\"" + p_v + "\"" + this.MakeTextOnChangeScript(rName, iName, p_v, bkCol) +">";
				break;
			  case "e": // Enum
//				pPn = pPn + "Enum : " + iName;
				pPn = pPn + "<SELECT id=\"" + iName + "\"" + this.MakeTextOnChangeScript(rName, iName, p_v, bkCol) +">";
				var j;
				var pTyr = pTy.r;
				for(j = 0; j < pTyr.length; j++) {
					pPn = pPn + "<OPTION value=\"" + pTyr[j].op + "\" title=\"" + pTyr[j].tt + "\"" + (pTyr[j].op == p_v ? " selected" : "") + ">" + pTyr[j].l + "</OPTION>";
				}
				pPn = pPn + "</SELECT>";
				break;
			  case "g": // Group
//				pPn = pPn + "Group: " + iName + ", " + pTy.r;
				pPn = pPn + "<TABLE>";
				pPn = pPn + this.MakeElementPropertyTable(pTy.r, preN + "_" + i, p_v);
				pPn = pPn + "</TABLE>";
				break;
			  default:
				pPn = pPn + "Type [" + pTy.t + "] unknown : " + iName;
				break;
			}
			pPn = pPn + "</TD></TR>";
		}
		return pPn;
	}	

	ShowPropertyDialog() {
		var draw = this.DN4.draw;
		var selEl = draw.currentSelection[0];
		var selElTdef = elementTypes[selEl.t];
		var fPs = selElTdef.p;
		var eName = selEl.n;
		
		$("#dialog_items").remove();
		var pPn = "<DIV id=\"dialog_items\"><TABLE>";
		pPn = pPn + "<TR title=\"Element ID (automatically assigned)\"><TH>ID</TH><TD ALIGN=\"RIGHT\">" + selEl.id + "</TD></TR>";
		pPn = pPn + "<TR title=\"Element type: " + selElTdef.tt + "\"><TH>Type</TH><TD ALIGN=\"RIGHT\">" + selElTdef.l + "</TD></TR>";
		pPn = pPn + "<TR title=\"Element Name\" id=\"rid_d_i_name\"><TH>Name</TH><TD ALIGN=\"RIGHT\"><INPUT type=\"text\" id=\"d_i_name\" value=\"" + eName + "\"" + this.MakeNameOnChangeScript("rid_d_i_name", "d_i_name", eName, "") +"></TD></TR>";
//		pPn = pPn + "<TR title=\"Number of properties\"><TH>#Props.</TH><TD ALIGN=\"RIGHT\">" + fPs.length + "</TD></TR>";
//		pPn = pPn + "<TR title=\"Name onchange script\"><TH>#Props.</TH><TD ALIGN=\"RIGHT\">" + this.MakeTextOnChangeScript("rid_d_i_name", "d_i_name", eName) + "</TD></TR>";
		pPn = pPn + "</TABLE>"; // End of common section with name and type

		// Real properties section
		pPn = pPn + "<DIV id=\"tabs\"><UL>";
		pPn = pPn + "<LI><A HREF=\"#F1\">Model</A></LI>";
		pPn = pPn + "<LI><A HREF=\"#F2\">Graph</A></LI>";
		pPn = pPn + "</UL>";
		pPn = pPn + "<DIV id=\"F1\"><TABLE><TR style=\"background-color: #cff;\"><TH>Property</TH><TH>Value</TH></TR>";
		pPn = pPn + this.MakeElementPropertyTable(fPs, "", selEl.p);
		pPn = pPn + "</TABLE></DIV>" // TAB1
		
		pPn = pPn + "<DIV id=\"F2\"><TABLE><TR style=\"background-color: #cff;\"><TH>Property</TH><TH>Value</TH></TR>";
		if(selElTdef.t == "n") {
			var gpnames = this.gpnames;
			for(i = 0; i < gpnames.length; i++) {
				var bkCol = (i % 2 == 0 ? "#ccc" : "#fff");
				var g_v = selEl.s[gpnames[i].n] * gpnames[i].s;
				pPn = pPn + "<TR title=\"" + gpnames[i].tt + "\" id=\"" + gpnames[i].ri + "\" style=\"background-color: " + bkCol + ";\"><TH>" + gpnames[i].l + "</TH><TD><INPUT type=\"text\" id=\"" + gpnames[i].fi + "\" value=\"" + g_v + "\"" + this.MakeTextOnChangeScript(gpnames[i].ri, gpnames[i].fi, g_v, bkCol) +"</TD></TR>";
			}
		}
		pPn = pPn + "</TABLE></DIV>"; // TAB2
		pPn = pPn + "</DIV>"; // TABs

		pPn = pPn + "</DIV>"; // DIALOG
//		$("#dialog").append(pPn);
		$("#dialog").append(pPn).data('DN4_model',this);

		dialogChangeLog = [];
		
		$( function() { $( "#dialog" ).dialog({title: "Properties of " + eName,
				width: 500,
				height: 450,
				buttons: {		// Apply and Revert seem too complex to create, and not worth
								// implementation has been delayed!
								//
/*			"Apply": function(){
			// update values, but keep dialog open
				var DN4_model = $("#dialog").data('DN4_model');
				DN4_model.UpdateProperties(true);
			},
			"Revert": function(){
			// restores values when dialog was opened
				var DN4_model = $("#dialog").data('DN4_model');
				DN4_model.RevertToPrevious();
			},*/
			 "Cancel": function(){$(this).dialog('close');} ,
			 "OK": function(){
			// update values
				var DN4_model = $("#dialog").data('DN4_model');
				DN4_model.UpdateProperties(false);
				DN4_model.DN4.draw.drawView();
				$(this).dialog('close');
			}
			 }});
			$("#tabs").tabs({heightStyle : "auto"});
	 
			 } );
	}
};